# This is the runtime Ruby definitions for PArrays.  

# to load the native part of the runtime.  When $IKRA_BACKEND is
# specified, say "cuda", it loads from "parray_cuda.so".  Otherwise,
# it loads from "parray_c.so".
require("parray_#{$IKRA_BACKEND || "c"}#{ENV['IKRA_DEBUG'] ? "_debug" : ""}")
IKRA_BACKEND=$IKRA_BACKEND || "c"

class PArray
  # to release the native memory allocated by PArray.alloc_int_array.
  # This must be defined as a class method so as not to reference the
  # object from the finalizer.
  def self.finalize(ptr)
    proc { PArray.free_array(ptr) }
  end

  # to allocate an array of size integers with a possibility of
  # invoking a garbage collector when it is failed to allocate on the
  # device memory.
  def self.alloc_int_array2(size, *bytes)
    PArray.alloc_int_array(size, *bytes) or begin
      GC.start
      PArray.alloc_int_array(size, *bytes) or begin
        throw NoMemoryError.new("in PArray.alloc_int_array")
      end
    end
  end

  # PArray.new(size) constructs a new parallel array object of the
  # given size elements.  
  def initialize(size, *bytes, &block)
    @size = size
    @ptr = PArray.alloc_int_array2(size, bytes ? bytes[0] : nil)
    # Register a finalizer to free the allocated pointer.  The
    # procedure object is created in a class method to avoid having a
    # reference from the finalizer to this object. See the following
    # page for the explanation.
    # http://www.mikeperham.com/2010/02/24/the-trouble-with-ruby-finalizers/
    ObjectSpace.define_finalizer(self, self.class.finalize(@ptr))
    if block
      yield(@ptr,@size)
    end
  end

  # self.reduce(zero){ |x,y| e(x,y) } performs reduction over the
  # elements in self by applying the given expression in the
  # accompanied block.  Unlike Array::reduce in pure Ruby, the order
  # of computation is not specified.  It only guarantees that
  #    [v0,..,vn].reduce(zero){ |x,y| e(x,y) } ==
  #    e([v0,  ..,vm].reduce(zero){|x,y|e(x,y)},
  #      [vm+1,..,vn].reduce(zero){|x,y|e(x,y)}).
  def reduce(*zero)
    yield(@ptr, *zero, @size)
  end

  # self.map{ |x| e(x) } performs e(x) for each element x in self, and
  # returns a new PArray of the same size that containing the results
  # of e(x).
  def map()
    destination = PArray.new(@size){} # empty block
    yield(@ptr, destination._get_ptr(), @size)
    destination
  end

  def map_zipped(tuple_size)
    elements = Array.new(tuple_size){ PArray.new(@size){} }
    destination = Ikra_Zipper.new([@size],*elements)
    yield(@ptr, *destination._get_ptrs(), @size)
    destination
  end

  # self.zip(a1,..,aM) creates a zipper object Z where
  # Z[indices] == [self[indices],a1[indices],...,aM[indices]]
  def zip(*arrays)
    Ikra_Zipper.new([@size], self, *arrays)
  end

  def neighbor3(pad)
    Ikra_Neighbor.new(self,pad)
  end
  def neighbor9(pad)
    Ikra_Neighbor.new(self,pad)
  end
  def neighbor27(pad)
    Ikra_Neighbor.new(self,pad)
  end

  # self.__get_ptr() returns a wrapper pointer of its internal
  # representation.  NOTE: It is unsafe to leak the pointer outside of
  # the object.
  def _get_ptr()
    @ptr
  end

  def _transfer_ptr()
    ObjectSpace.undefine_finalizer(self)
    p = @ptr
    @ptr = nil
    p
  end

  def self.new_(*sizes,&block)
    PArrayM.new(*sizes,&block)
  end

  def _get_str()
    PArray.int_array2str(_get_ptr(), @size)
  end

  def sizes()
    [@size]
  end

  def pack(format)
    if format=="I!*"
      PArray.int_array2str(@ptr, @size)
    else
      throw Exception.new("unsupported format #{format}")
    end
  end

  # TODO: support multi dimensional arrays

  def self.ruby_to_i1(ruby_array)
    len = ruby_array.length
    bytes = ruby_array.pack("I!*")
    PArray.new(len,bytes)
  end

  def self.ruby_to_f1(ruby_array)
    len = ruby_array.length
    bytes = ruby_array.pack("f*")
    PArray.new(len,bytes)
  end

  def self.array_size_of(ruby_array, dim)
    if dim==0
      ruby_array.length
    else
      ruby_array.map{ |array_sub| 
        self.array_size_of(array_sub, dim-1)
      }.max
    end
  end

  def self.array_ref(ruby_array, sizes, indices, default)
    sizes.zip(indices).reduce( ruby_array ){ | array, (size,index) |
      if index < array.length then array[index] else return default end
    }
  end

  def self.indices(rest_sizes, i)
    indices,_ = rest_sizes.reduce( [ [], i ] ){ | (indices, ii), rest_size |
      [ indices + [ii / rest_size], ii % rest_size ]
    }
    indices
  end

  def self.ruby_to_iN(ruby_array, rank)
    # (max) sizes of each dimension
    sizes = Array.new(rank){ |dim| self.array_size_of(ruby_array, dim) }
    # the number of elements of the trailing dimensions
    total_size, *rest_sizes = sizes.reverse.
      reduce( [1] ){ | (s,*sizes) , dim | [dim*s, s, *sizes] }
    # a flat array of elements in ruby_array; when ruby_array is irregular,
    # 0 is padded.
    elements = Array.new(total_size){ |i|
      self.array_ref(ruby_array, sizes, self.indices(rest_sizes, i),  0)
    }
    PArrayM.new(PArray.alloc_int_array(total_size, elements.pack("I!*")),
                *sizes){}
  end

  def self.ruby_to_fN(ruby_array, rank)
    # (max) sizes of each dimension
    sizes = Array.new(rank){ |dim| self.array_size_of(ruby_array, dim) }
    # the number of elements of the trailing dimensions
    total_size, *rest_sizes = sizes.reverse.
      reduce( [1] ){ | (s,*sizes) , dim | [dim*s, s, *sizes] }
    # a flat array of elements in ruby_array; when ruby_array is irregular,
    # 0 is padded.
    elements = Array.new(total_size){ |i|
      self.array_ref(ruby_array, sizes, self.indices(rest_sizes, i),  0)
    }
    PArrayM.new(PArray.alloc_int_array(total_size, elements.pack("f*")),
                *sizes){}
  end

end

class Array
  def to_a_i()  PArray.ruby_to_i1(self) end
  def to_a_i1() PArray.ruby_to_i1(self) end
  def to_a_i2() PArray.ruby_to_iN(self, 2) end
  def to_a_i3() PArray.ruby_to_iN(self, 3) end
  def to_a_i4() PArray.ruby_to_iN(self, 4) end
  def to_a_i5() PArray.ruby_to_iN(self, 5) end
  def to_a_i6() PArray.ruby_to_iN(self, 6) end
  def to_a_i7() PArray.ruby_to_iN(self, 7) end
  def to_a_i8() PArray.ruby_to_iN(self, 8) end
  def to_a_i9() PArray.ruby_to_iN(self, 9) end
  def to_a_f()  PArray.ruby_to_f1(self) end
  def to_a_f1() PArray.ruby_to_f1(self) end
  def to_a_f2() PArray.ruby_to_fN(self, 2) end
  def to_a_f3() PArray.ruby_to_fN(self, 3) end
  def to_a_f4() PArray.ruby_to_fN(self, 4) end
  def to_a_f5() PArray.ruby_to_fN(self, 5) end
  def to_a_f6() PArray.ruby_to_fN(self, 6) end
  def to_a_f7() PArray.ruby_to_fN(self, 7) end
  def to_a_f8() PArray.ruby_to_fN(self, 8) end
  def to_a_f9() PArray.ruby_to_fN(self, 9) end
end

class PArrayM
  # multi-dimensional arrays

  # to release the native memory allocated by PArray.alloc_int_array.
  # This must be defined as a class method so as not to reference the
  # object from the finalizer.
  def self.finalize(ptr)
    proc { PArray.free_array(ptr) }
  end

  # PArray.new(size) constructs a new parallel array object of the
  # given size elements.  
  def initialize(*sizes)
    case sizes[0]
    when Integer
      @sizes = sizes
      @rank = sizes.length
      @ptr = PArray.alloc_int_array2(sizes.reduce(&:*),nil)
    when Object
      @ptr = sizes.shift
      @sizes = sizes
      @rank = sizes.length
    end      
    # Register a finalizer to free the allocated pointer.  The
    # procedure object is created in a class method to avoid having a
    # reference from the finalizer to this object. See the following
    # page for the explanation.
    # http://www.mikeperham.com/2010/02/24/the-trouble-with-ruby-finalizers/
    ObjectSpace.define_finalizer(self, self.class.finalize(@ptr))
    yield(@ptr,*@sizes)
  end

  # self.reduce(zero){ |x,y| e(x,y) } performs reduction over the
  # elements in self by applying the given expression in the
  # accompanied block.  Unlike Array::reduce in pure Ruby, the order
  # of computation is not specified.  It only guarantees that
  #    [v0,..,vn].reduce(zero){ |x,y| e(x,y) } ==
  #    e([v0,  ..,vm].reduce(zero){|x,y|e(x,y)},
  #      [vm+1,..,vn].reduce(zero){|x,y|e(x,y)}).
  def reduce(*zero)
    yield(@ptr, *zero, *@sizes)
  end

  # self.map{ |x| e(x) } performs e(x) for each element x in self, and
  # returns a new PArray of the same size that containing the results
  # of e(x).
  def map()
    destination = PArrayM.new(*@sizes){} # empty block
    yield(@ptr, destination._get_ptr(), *@sizes)
    destination
  end

  def map_r(other)
    destination = PArrayM.new_(other._transfer_ptr, *@sizes){} # empty block
    yield(@ptr, destination._get_ptr(), *@sizes)
    destination
  end



  # self.zip(a1,..,aM) creates a zipper object Z where
  # Z[indices] == [self[indices],a1[indices],...,aM[indices]]
  def zip(*arrays)
    Ikra_Zipper.new(@sizes, self, *arrays)
  end

  def neighbor3(pad)
    Ikra_Neighbor.new(self,pad)
  end
  def neighbor9(pad)
    Ikra_Neighbor.new(self,pad)
  end
  def neighbor27(pad)
    Ikra_Neighbor.new(self,pad)
  end

  def neighborE3()
    Ikra_NeighborE.new(self)
  end
  def neighborE9()
    Ikra_NeighborE.new(self)
  end
  def neighborE27()
    Ikra_NeighborE.new(self)
  end
  
  # self.__get_ptr() returns a wrapper pointer of its internal
  # representation.  NOTE: It is unsafe to leak the pointer outside of
  # the object.
  def _get_ptr()
    @ptr
  end
  def _transfer_ptr()
    ObjectSpace.undefine_finalizer(self)
    p = @ptr
    @ptr = nil
    p
  end

  def sizes()
    @sizes
  end

  def pack(format)
    if format=="I!*"
      PArray.int_array2str(@ptr, @sizes.reduce(&:*))
    else
      throw Exception.new("unsupported format #{format}")
    end
  end
end

# to represent an array of tuples
class Ikra_Zipper

  def initialize(sizes, *arrays)
#    puts "Ikra_Zipper with #{arrays.length} arrays"
    @sizes = sizes
    @arrays = arrays
  end

  def self.new_alloc(sizes, num_arrays, &initializer)
    if sizes.length==1
      as = Array.new(num_arrays){ PArray.new(sizes[0]) }
    else
      as = Array.new(num_arrays){ PArrayM.new(sizes[0]) }
    end
    z = Ikra_Zipper.new(sizes, *as)
    z.initialize_elements(&initializer)
    z
  end

  # to obtain a flat list of array pointers.
  # Ikra_Zipper.new(s,a1,a2)._get_ptrs = [a1._get_ptr, a2._get_ptr]
  # Ikra_Zipper.new(s,Ikra_Zipper.new(s,a1,a2),Ikra_Zipper.new(s,a3)).
  #     _get_ptrs = [a1._get_ptr, a2._get_ptr, a3._get_ptr]
  def _get_ptrs()
    @arrays.map{|a| a._get_ptr}.flatten
  end

  def initialize_elements()
    yield(*self._get_ptrs(), *@sizes)
    self
  end

  def map()
    destination = PArray.new_(*@sizes){}
    yield(*self._get_ptrs, destination._get_ptr(), *@sizes)
    destination
  end

  def map_zipped(tuple_size)
    raise "multi-dimensional arrays are not yet supported" unless
      @sizes.length == 1
    elements = Array.new(tuple_size){ PArray.new(@sizes[0]){} }
    destination = Ikra_Zipper.new(@sizes,*elements)
    yield(*self._get_ptrs, *destination._get_ptrs(), *@sizes)
    destination
  end

  def zip(*arrays)
    Ikra_Zipper.new(@sizes, *@arrays, *arrays)
  end

end


class Ikra_Neighbor
  def initialize(original,pad)
    @original = original
    @pad = pad
  end

  def map()
    sizes = @original.sizes()
    destination = PArray.new_(*sizes){}
    yield(@original._get_ptr(), @pad, destination._get_ptr(), *sizes)
    destination
  end
end

class Ikra_NeighborE
  def initialize(original)
    @original = original
  end

  def map()
    sizes = @original.sizes()
    destination = PArray.new_(*sizes){}
    yield(@original._get_ptr(), 0, destination._get_ptr(), *sizes)
    destination
  end
  
  def map_r(victim)
    sizes = @original.sizes()
    destination = PArray.new_(victim._transfer_ptr,*sizes){}
    yield(@original._get_ptr(), 0, destination._get_ptr(), *sizes)
    destination
  end

end
