(* Reference to the Ruby's syntax
   http://en.wikibooks.org/wiki/Ruby_Programming/Syntax
*)

type variable = string

type type_decl = variable * Ikra_type.t

let string_of_type_decls : ('a -> string) -> type_decl list -> string =
  fun string_of_type ->  function
    | [] -> ""
    | decls ->
      "_typedecl(" ^ 
	(String.concat ","
	   (List.map (fun (v,t) -> (string_of_type t) ^ ",:" ^ v) decls)) ^
	");"

type parameter_pattern =
  | Var_pattern of string * Ikra_type.t
  | Array_pattern of parameter_pattern list

let empty_pattern = Array_pattern []

type literal =
  | Integer_literal of int
  | Float_literal of float
  | String_literal of string
  | Symbol_literal of string
  | Range_literal of int * int		(* due to the parser *)

type ast_body =
  (* method call (receiver, method_name, operands, block) *)
  (* NOTE: the receiver is now mandatory; if ommitted in the source,
     it should be "self" *)
  | Call of tast * string * tast list * tast option
  (* literal (literal) *)
  | Literal of literal
  (* assignment to a constant (name,tast) *)
  | Constant_assign of string * tast
  (* constant (name) *)
  | Constant of string 
  (* assignment to local variable (name,tast) *)
  | Local_assign of string * tast
  (* local variable reference (name) *)
  | Local_var of string
  (* block structure (variables, body)*)
  | Block of tast list
  (* procedure (as a parameter) *)
  | Proc of parameter_pattern * tast
  (* method definition (method_name, variables, body) *)
  | Method_def of string * parameter_pattern * tast
  (* yield (tast) *)
  | Yield of tast
  (* array literal (tast list) *)
  | Array of tast list
  (* while loop (cond, body) *)
  | While of tast * tast 
  (* conditional branch (cond, then, else) *)
  | If of tast * tast * tast option
  (* for loop (var, range, body) *)
  | For of string * tast * tast
  (* return from a function (exp1) *)
  | Return of tast option
  (* type_decl (tdecl list, body) --- only after type inference *)
  | Type_decl of type_decl list * tast list
and tast = ast_body * Ikra_type.t

let is_constant : tast -> bool = function
  | Constant _,_ -> true
  | _ -> false

let is_constant_def : string -> tast -> bool =
   (* to determine if the given expression is an assignment to the
      given constant. *)
  fun cname -> function
    | Constant_assign(cname',_),_ -> cname=cname'
    | _                          -> false

let is_method_def : string -> tast -> bool =
  (* is_method_def mname exp = true if exp is a method defintion of
     the given mname *)
  fun mname -> function
    |Method_def(mname',_,_),_ ->mname=mname'
    | _->false
let method_def_body : tast -> tast = function
  | Method_def(_,_,body),_ -> body
  | _ -> invalid_arg "not a method definition"


module Untyped_constructor = struct
  open Ikra_type
  let var_pattern name = Var_pattern(name,Tobject)
  let array_pattern patterns = Array_pattern(patterns)
  let empty_pattern = array_pattern[]

  let call (receiver, method_name, operands, block) =
    (Call(receiver,method_name,operands,block),Tobject)
  let literal lit = (Literal(lit),Tobject)
  let constant_assign (name,exp) = Constant_assign(name,exp),Tobject
  let constant name = (Constant(name),Tobject)
  let local_assign (name,exp) = Local_assign(name,exp),Tobject
  let local_var name = Local_var name,Tobject
  let block (body) = Block(body),Tobject
  let proc (variables,body) = Proc(variables,body),Tobject
  let method_def (method_name, variables, body) =
    Method_def(method_name, variables, body),Tobject
  let yield exp = Yield exp,Tobject
  let array exps = Array exps, Tobject
  let while_ (cond,body) = While(cond,body),Tobject
  let if_ (cond,then_exp,else_exp) = If(cond,then_exp,else_exp),Tobject
  let for_ (var,range,body) = For(var,range,body),Tobject
  let return exp = Return (Some exp),Tobject
  let return0 () = Return None,Tobject
  let return_o exp = Return exp,Tobject
  let type_decl (decls, body) = Type_decl(decls,body),Tobject
end
  
let pseudo_infix_alist : (string * string) list =
  ["__PSEUDO_dot2", "..";
   "not@",          "!";
     "&&",          "&&";
     "||",          "||";
  ]

(* When a parser converts a method call without receiver, it uses the
   implicit self as the receiver.  The variable shall be deleted when
   it is unparsed.  We use a different name from "self" because "m()"
   and "self.m()" are not equivalent in Ruby when m is private. *)
let implicit_self_name = "**self**"
let implicit_self = Untyped_constructor.local_var implicit_self_name
let is_implicit_self : tast -> bool = function
  | (Local_var name,_) -> name = implicit_self_name
  | _ -> false
let is_self : tast -> bool = function
  | (Local_var name,_) -> name = implicit_self_name || name = "self"
  | _ -> false

let pseudo_infix_revalist : (string * string) list =
  List.map (fun (a,b) -> b,a) pseudo_infix_alist
let pseudo_name_of : string -> string =
  fun name -> List.assoc name pseudo_infix_revalist
let is_pseudo_infix : string -> bool =
  fun name -> List.mem_assoc name pseudo_infix_alist
let pseudo_infix : string -> string =
  fun name -> List.assoc name pseudo_infix_alist

let in_paren = Util.in_paren

let map_concat : string -> ('a -> string) -> 'a list -> string =
  (* to convert elements in the given lst by using the given to_str,
     and then concatenate by separating with the given sep. *)

  fun sep to_str lst -> String.concat sep (List.map to_str lst)

let string_of_literal : literal -> string =
  function
    | Integer_literal n -> string_of_int n
    | Float_literal n -> Printf.sprintf "%f" n (* do not use
						  string_of_float
						  here, as it omits
						  fractions when they
						  are zero. *)
    | String_literal s -> "\"" ^ s ^ "\""
    | Symbol_literal s -> ":" ^ s
    | Range_literal _ ->
      raise (Invalid_argument "An AST should not contain a Range_literal \
			       because it must have been converted into a \
			       call expression with a pseudo method name.")

let rec string_of_block_parameter_sub
    : ?show_type:bool -> parameter_pattern -> string =
  fun ?(show_type = false) ->
  function
    | Var_pattern (var_name, typ)
      -> var_name ^ if show_type then ":"^(Ikra_type.string_of typ) else ""
    | Array_pattern patterns
      -> "(" ^ (map_concat ","
		  (string_of_block_parameter_sub ~show_type)
		  patterns) ^ ")" 

let string_of_block_parameter : ?show_type:bool -> parameter_pattern -> string =
  fun ?(show_type = false) ->
  function
    | Array_pattern [] -> ""
    | Array_pattern patterns ->
      " | " ^ (map_concat ","
		 (string_of_block_parameter_sub ~show_type)
		 patterns) ^ " | "
    | Var_pattern _ as vp ->
      " | " ^ (string_of_block_parameter_sub ~show_type vp) ^ " | "

let rec string_of : ?hide_type_decl:bool -> ?show_type:bool -> tast -> string =
  (* to compute a string representation of an AST *)
  fun ?(hide_type_decl = false) ?(show_type = false) (exp,typ) ->
    let str_of = string_of  ~hide_type_decl ~show_type in
    let maybe_string_of_exp = fun prefix maybe_exp suffix ->
      match maybe_exp with
	| None -> ""
	| Some exp -> prefix ^ (str_of exp) ^ suffix in
    let s = match exp with
      | Constant n -> n 
      | Literal l -> string_of_literal l
      | Constant_assign(var,exp) | Local_assign(var,exp)
        -> var ^ " = " ^ (in_paren (str_of exp))
      | Local_var(var) -> var
      (* special binary operators like e1..e2 *)
      | Call(receiver, method_name, [e1;e2], None)
      	  when is_implicit_self(receiver) && is_pseudo_infix method_name ->
	(in_paren (str_of e1))
	^ (pseudo_infix method_name)
	^ (in_paren (str_of e2))
      | Call(receiver, method_name, [e1], None)
      	  when is_implicit_self(receiver) && is_pseudo_infix method_name ->
	(pseudo_infix method_name)
	^ (in_paren (str_of e1))
      | Call(receiver, method_name, [e1], None)
      	  when is_pseudo_infix method_name ->
        (in_paren (str_of receiver))
	^ (pseudo_infix method_name)
	^ (in_paren (str_of e1))
      | Call(receiver, method_name, operands, block) ->
	(if is_implicit_self receiver
	 then "" else (str_of receiver) ^ ".")
	^ method_name
	^ (in_paren (map_concat "," str_of operands))
	^ (maybe_string_of_exp "{" block "}")
      | Block(body) -> (map_concat ";" str_of body)
      | Proc(params,body) ->
	(string_of_block_parameter ~show_type params)
	^ (str_of body)
      | Method_def(method_name, variables, body) ->
	"def " ^ method_name ^ (string_of_block_parameter_sub
				  ~show_type variables) ^
	  (str_of body) ^ "; end" 
      | Yield exp -> "yield" ^ (in_paren (str_of exp))
      | Array exps -> "[" ^ (map_concat "," str_of exps) ^ "]"
      | While(cond, body) ->
	"while " ^ (str_of cond) ^ ";" ^ (str_of body) ^ "; end"
      | If(cond, then_exp, else_exp) ->
	"if " ^ (str_of cond) ^ " then " ^ (str_of then_exp)
	^ (maybe_string_of_exp " else " else_exp "") ^ " end"
      | For(var, range, body) ->
	"for " ^ var ^ " in " ^ (str_of range) ^ " do " ^
	  (str_of body) ^ " end"
      | Return(Some exp1) -> "return " ^ (str_of exp1)
      | Return(None)      -> "return"
       (* only in a typed program *)
      | Type_decl(tdecls, body) when not(hide_type_decl) ->
	(string_of_type_decls Ikra_type.string_of tdecls) ^
	  (map_concat ";" str_of body)
      | Type_decl(tdecls, body) (* when hide_type_decl *) ->
	(map_concat ";" str_of body)
    in if show_type
      then Printf.sprintf "(%s):%s" s (Ikra_type.string_of typ)
      else s

let mlexp_of : ('a -> string) -> tast -> string =
  (* mlexp_of type_printer exp constructs an OCaml expression of exp
     as a string.  type_printer is a helper function that converts a type
     expression into a string. *)

  fun type_printer ->
    let mlexp_of_literal = function
      | Integer_literal n -> Printf.sprintf "Integer_literal %d" n
      | Float_literal n -> Printf.sprintf "Float_literal %f" n
      | String_literal s -> Printf.sprintf "String_literal \"%s\"" s
      | Symbol_literal s -> Printf.sprintf "Symbol_literal \"%s\"" s
      | Range_literal(f,t) -> Printf.sprintf "Range_literal(%d,%d)" f t in

    let maybe f = function
      | Some n -> Printf.sprintf "Some(%s)" (f n)
      | None -> "None" in

    let map f l = "["^ (String.concat ";" (List.map f l)) ^ "]" in
    
    let rec mlexp_of_parameter_pattern = function
      | Var_pattern (s,_) -> Printf.sprintf "Var_pattern(\"%s\")" s
      | Array_pattern l -> Printf.sprintf "Array_pattern%s"
	(map mlexp_of_parameter_pattern l) in

    let mlexp_of_typedecl type_printer (v,t)= 
      "(\"" ^ v ^ "\"," ^ (type_printer t) ^ ")" in

    let rec exp_of (exp,_) =
      match exp with
	| Constant_assign(var,exp) -> 
	  Printf.sprintf("Constant_assign(\"%s\",%s)") var (exp_of exp)
	| Constant n -> Printf.sprintf("Constant(\"%s\")") n 
	| Literal l -> Printf.sprintf("Literal(%s)") (mlexp_of_literal l)
	| Local_assign(var,exp) -> 
	  Printf.sprintf("Local_assign(\"%s\",%s)") var (exp_of exp)
	| Local_var(var) -> Printf.sprintf("Local_var(\"%s\")") var
	| Call(receiver, method_name, args, block) ->
	  Printf.sprintf "Call(%s,\"%s\",%s,%s)"
	    (exp_of receiver) method_name
	    (map exp_of args) (maybe exp_of block)
	| Block(body) -> Printf.sprintf("Block%s") (map exp_of body)
	| Proc(params,body) ->
	  Printf.sprintf("Proc(%s,%s)")
	    (mlexp_of_parameter_pattern params) (exp_of body)
	| Method_def(method_name, variables, body) ->
	  Printf.sprintf("Method_def(\"%s\", %s, %s)")
	    method_name
	    (mlexp_of_parameter_pattern variables)
	    (exp_of body)
	| Yield exp -> Printf.sprintf "Yield(%s)"  (exp_of exp)
	| Array exps -> Printf.sprintf "Array%s" (map exp_of exps)
	| While(cond, body) -> Printf.sprintf "While(%s,%s)"
	  (exp_of cond) (exp_of body)
	| If(cond, then_exp, else_exp) ->
	  Printf.sprintf "If(%s,%s,%s)" 
	    (exp_of cond) (exp_of then_exp) (maybe exp_of else_exp)
	| For(var, range, body) ->
	  Printf.sprintf "For(\"%s\",%s,%s)" 
	    var  (exp_of range) (exp_of body)
        | Return(exp1) -> Printf.sprintf "Return(%s)" (maybe exp_of exp1)
	(* only in a typed program *)
	| Type_decl(tdecls, body) ->
	  Printf.sprintf "Type_decl(%s,%s)" 
	    (map (mlexp_of_typedecl type_printer) tdecls)
	    (map exp_of body) in
    exp_of

let rec variables_in_pattern : parameter_pattern -> string list =
  (* to collect all the variables in the given pattern *)

  function
    | Var_pattern (v,_) -> [v]
    | Array_pattern sub_patterns ->
      List.concat (List.map variables_in_pattern sub_patterns)

module VarSet = Set.Make(String)

let string_of_varset : VarSet.t -> string =
  (* to compute a string representation of the given set *)
  
  fun set ->
    "{" ^(String.concat "," (VarSet.elements set))^"}"

let rec variable_set_in_pattern : parameter_pattern -> VarSet.t =
  (* to compute a set of variables declared in the given params
     pattern. *)

  fun params -> 
    List.fold_left (fun set elt -> VarSet.add elt set)
      VarSet.empty (variables_in_pattern params)

let map_union : ('a -> VarSet.t) -> 'a list -> VarSet.t =
  (* apply the given function f to all the elements in lst, and
     returns a union of sets return from the applications. *)
  
  fun f lst -> 
    List.fold_left (fun set elt -> VarSet.union set (f elt))
      VarSet.empty lst

let rec free_variables : tast -> VarSet.t =
  (* to compute a set of free variables in the gvien expression *)
  (* let free_variables (exp,_) = free_variables exp in *)
  fun ((exp,_) as texp) ->
  match exp with
    (* fv <<e1.m(e2,..en){ eb }>> = fv(e1)+...+fv(en)+fv(eb) *)
    | Call(reciever,_,args,block) ->
      VarSet.union
	(VarSet.union
	   (free_variables reciever)
	   (map_union free_variables args))
	(maybe_free_variables block)
    | Literal _ -> VarSet.empty
    (* fv <<C=e>> = fv(e) *)
    | Constant_assign(var,exp) -> free_variables exp
    (* fv <<Object>> = {  } *)
    (* NOTE: constants are not free variables as they don't need to be
       passed as lexical parameters *)
    | Constant cname -> VarSet.empty
    (* fv <<v=e>> = { v } + fv(e) *)
    | Local_assign(var,exp) -> VarSet.add var (free_variables exp)
    (* fv <<self>> = { } *)
    | Local_var var when is_self texp -> VarSet.empty
    (* fv <<v>> = { v } *)
    | Local_var var -> VarSet.singleton var
    (* fv <<{ e1;..;em }>> = fv(e1)+..+fv(em) *)
    | Block(exps) -> (map_union free_variables exps)
    (* fv <<{ |v1,..vn| e }>> = fv(e)\{v1,..,vn} *)
    | Proc(params,exp) ->
      VarSet.diff (free_variables exp) (variable_set_in_pattern params)
    (* fv <<def m(v1,..vn) e1;..em end>> = (fv(e1)+..+fv(em))\{v1,..,vn} *)
    | Method_def(_,params,body) ->
      VarSet.diff (free_variables body) (variable_set_in_pattern params)
    (* fv <<yield e>> = fv(e) *)
    | Yield exp -> free_variables exp
    (* fv <<[e1,..,en]>> =  fv(e1)+..+fv(en) *)
    | Array exps -> map_union free_variables exps
    (* fv <<while e0 do e1;..;en end>> =  fv(e0)+..+fv(en) *)
    | While(cond,exp) ->
      VarSet.union (free_variables cond) (free_variables exp)
    (* fv <<if e0 then e1 else e2 end>> =  fv(e0)+fv(e1)+fv(e2) *)
    | If(cond,then_exp,else_exp) ->
      VarSet.union (VarSet.union (free_variables cond)
		      (free_variables then_exp))
	(maybe_free_variables else_exp)
    (* fv <<for v in e0 do e1 end>> =  fv(e0)+(fv(e1)\{v}) *)
    | For(var,range,body) ->
      VarSet.union(free_variables range)
	(VarSet.remove var (free_variables body))
    (* fv <<return e1>> = fv(e1) *)
    | Return(exp1) -> maybe_free_variables exp1

    (* fv <<_typedecl(t0,v0,...,tn,vn); e>> -> fv(e)\{v1,..,vn}  *)
    | Type_decl (decl,exps) ->
      VarSet.diff (map_union free_variables exps)
	(List.fold_right (fun (v,t) set -> VarSet.add v set) decl VarSet.empty)

and maybe_free_variables : tast option -> VarSet.t =
  (* to compute a set of free variables in the given (maybe)
     expression *)
  
  function
    | None -> VarSet.empty
    | Some exp -> free_variables exp


let rec recognize_paratemer_pattern :
    parameter_pattern -> parameter_pattern = function
    | Var_pattern (var, t) -> Var_pattern (var, t(* Ikra_type.Tobject *))
    | Array_pattern parameters ->
      Array_pattern (List.map recognize_paratemer_pattern parameters)


let rec ikra_type_of_ast : tast -> Ikra_type.t =
  (* of_ast expression constructs an Ikra type representation for a
     type expressed as a Ruby expression*)
  function
    | Constant("Integer"),_ -> Ikra_type.Tint
    | Constant("Float"),_ -> Ikra_type.Tfloat
    | Constant(_),_ -> Ikra_type.Tobject	(* others are dynamic object *)
    | Array [Constant("Array"),_;
	     Literal(Integer_literal n),_;
	     t],_ -> Ikra_type.Tarray(n, ikra_type_of_ast t)
    | Array [Constant("Neighbor"),_;
	     Literal(Integer_literal m),_;
	     Literal(Integer_literal n),_;
	     t],_ -> Ikra_type.Tneighbor(m, n, ikra_type_of_ast t)
    | Array ((Constant("Tuple"),_)::element_types),_ ->
      Ikra_type.Ttuple(List.map ikra_type_of_ast element_types)
    | e -> raise (Invalid_argument
		    ((string_of e) ^
			" cannot be recognized as a type."))

let is_ikra_type : tast -> bool =
  (* is_ikra_type exp determines if exp can be recognized as an Ikra
     type *)

  fun exp -> 
    try let _ = ikra_type_of_ast exp in true
    with Invalid_argument _ -> false


let remove_first : ('a -> bool) -> 'a list -> ('a * 'a list) =
  fun test lst ->
    let rec loop = function
      | ([],_) -> raise Not_found
      | (hd::tl,prev) when test hd -> (hd, (List.rev prev) @ tl)
      | (hd::tl, prev) -> loop (tl, hd::prev) in
    loop(lst,[])


let return_type_of_body : tast -> (Ikra_type.t * tast) option =
  function 
    | (Block((Type_decl(decls,body_exps),t)::rest_exps)),tt  ->
      (try let ((_,typ), other_decls) =
	     remove_first (fun (name, _) -> name = "_RETURN") decls in
	   Some(typ, ((Block((Type_decl(other_decls, body_exps),typ)
                      ::rest_exps)), tt))
       with Not_found -> None)
    | _ -> None

let lookup_in_tdecl :
    string -> Ikra_type.t -> type_decl list -> Ikra_type.t =
  fun var default decls ->
    try List.assoc var decls
    with Not_found -> default

let replace_types
    : tast -> parameter_pattern -> parameter_pattern =
  function
    | Block((Type_decl(decls, _),_)::_),_ ->
      let rec replace =
	function
	  | Var_pattern (v,t) -> Var_pattern(v, lookup_in_tdecl v t decls)
	  | Array_pattern pat -> Array_pattern (List.map replace pat) in
      replace
    | _ -> fun pat -> pat

let maybe = Util.Option.maybe 

let rec recognize : tast -> tast =
  (* recognize ast constructs a typed AST with recognizing Ruby
     encoded types *)

  let convert = function
    (* method call *)
    | Call (receiver, method_name, operands, block) ->
      Call (recognize receiver,
	    method_name,
	    List.map recognize operands,
	    maybe recognize block)
    (* literal  *)
    | Literal(literal) -> Literal literal
    (* assignment to a constant *)
    | Constant_assign (name,ast) -> Constant_assign (name,recognize ast)
    (* constant  *)
    | Constant(name) -> Constant(name)
    (* assignment to local variable *)
    | Local_assign (name,ast) -> Local_assign (name,recognize ast)
    (* local variable reference *)
    | Local_var (name) ->Local_var (name)
    (* block structure*)
    | Block (body) -> Block (List.map recognize body)
    (* procedure as a parameter *)
    | Proc (variables, body) ->
      Proc (recognize_paratemer_pattern variables, recognize body)
    (* method definition *)
    | Method_def (method_name, variables, body) ->
      Method_def (method_name,
		  recognize_paratemer_pattern variables,
		  recognize body)
    (* yield *)
    | Yield(exp) -> Yield(recognize exp)
    (* array literal *)
    | Array(exps) -> Array(List.map recognize exps)
    (* while loop *)
    | While (cond, body) -> While (recognize cond, recognize body)
    (* conditional branch *)
    | If (cond, then_exp, else_exp) ->
      If (recognize cond, recognize then_exp, maybe recognize else_exp)
    (* for loop *)
    | For (var, range, body) ->
      For (var, recognize range, recognize body)

    (* return *)
    (* TODO: should it have the type specified by surrounding
       type_decl? *)
    | Return(exp1) -> Return(maybe recognize exp1) 

    (* type_decl --- only after type inference *)
    | Type_decl (tdecls, body)
      -> Type_decl (tdecls, List.map recognize body) in

  fun (exp,t) -> match (convert exp) with
    (* Convert a pseudo function call to _t into a type of the
       expression.
       _t(exp,t) -> exp of type t
    *)
    | Call (rcv , "_t", [exp,_;texp], None) when is_implicit_self rcv ->
      exp, (ikra_type_of_ast texp)

    (* When a block has a _type_decl inside that specifies the type of
       _RETURN pesudo variable, use the type as the type of the block
       itself.
       { |v1,...| _type_decl(...,t, :_RETURN,...) ... } 
       -> { |v1,...| _type_decl(...,...) ... } of type t
    *)
    | Proc(variables, body)
        when Util.Option.is_some (return_type_of_body body) ->
      let typ,body_exp = Util.Option.get(return_type_of_body body) in
      let typed_vars = replace_types body variables in
      (Proc(typed_vars, body_exp), typ)
    | e ->  e,t
