(* convert a (partially) typed Ikra program into a (kernel invoking)
   Ruby program and a CUDA program. *)
module type CONVERT_KERNEL = sig
  type kind = Initialize | Map | Reduce | UnitlessReduce
  val string_of_kind : kind -> string
  val element_type_of : Ikra_type.t -> Ikra_type.t
  type kernel = {
    kind : kind;
    name : string;
    free_vars : Ast.tast list;
    params : Ast.parameter_pattern;
    type_decls : Ast.type_decl list;
    body : Ast.tast list;
    body_type : Ikra_type.t;
    receiver : Ast.tast;
    top_decls : Ast.tast;
    return_type : Ikra_type.t;
  }
  val top_decls : kernel list -> Ast.tast list
  type var_iname_type_cname = Ast.tast * string
  type env = {
    wrapper_name : string;
    kernel_fun_name : string;
    body_fun_name : string;
    argspecs : var_iname_type_cname list;
    body_ctype : Cuda.cuda_type;
    fvspecs : (Ast.tast * string) list;
    body_args : (Cuda.cuda_type * Cuda.variable) list;
    element_statements : Cuda.statement list;
  }
  val generate_makefile : generate_only:bool -> Convert_env.t -> unit
  val cuda_code_of : string -> kernel list -> string

  val cxx_suffix : string 
  val backend_name : string 
end

(* ********************************************************************** *)

module M (Convert_kernel : CONVERT_KERNEL) = struct

type tast = Ast.tast
type variable = Ast.variable
let tint = Ikra_type.Tint
let tobject = Ikra_type.Tobject

let target_ruby_suffix = "ikra.rb"
let target_cuda_suffix = "ikra." ^ Convert_kernel.cxx_suffix (* "cc" *)

type array_ops = Convert_kernel.kernel

let maybe = Util.Option.maybe

let is_of_array_type : variable -> Ikra_type.t Env.env -> bool =
  (* is_of_array_type v env determines if v has an array type in
     env *)

  fun v env -> match Env.lookup env v with
    | Ikra_type.Tarray _ -> true
    | _ -> false

let dup n name typ =
  let rec loop i =
    if i=n then []
    else (name ^ (string_of_int i), typ)::(loop (i+1)) in
  loop 0

let array_op : Ikra_type.t Env.env -> tast ->
  (Convert_kernel.kind * (string * Ikra_type.t) list * string * tast
   * tast list * tast) option =
  (* array_op env expression extracts 

     1. the type of the kernel function (either initialize, reduce,
     unitless_reduce or map),

     2. the list of variables that will be provided to the wrapper,

     3. the name of the method in the converted expression, 
     
     4. a replacement of the receiver expression if the expression is
     an array operation that shall be realized by the CUDA kernel,

     5. a replacement of the list of arguments, and 

     6. a replacement of the block experssion if the call has a block
     argument with &.  (e.g., o.m(&:f) then create a block expression
     like this: { |x,y| x.f(y) } *)

  let open Ast in

      let explicit_block = function
        (* given an expression, it returns a proper block expression
           as an Ruby AST.  When the given expression is already a
           block, the expression is returned as it is.  When the given
           expression is a symbol, it constructs a block expression
           like { |v1,...,vN| v1.sym(v2,..,vN) } where sym is the
           given symbol.  This is used for constructing a kernel for
           
           a.reduce(&:+) 

           which should be interpreted as

           a.reduce(){ |x,y| x+y }.

        *)

        (* when the expression is already a block *)
        | Proc _,_ as e -> e
        (* when the expression is a symbol, the type must tell the
           number of necessary arguments. *)
        | Literal(Symbol_literal operation),
          ( Ikra_type.Tproc(arg_types, return_type, None))  ->
          (* create formal parameters for the block *)
          let vars
              = Util.mapi (fun i t -> ("v"^(string_of_int i), t)) arg_types in
          let var_pattern = List.map (fun (v,t) -> Var_pattern(v,t)) vars in
          let var_exps = List.map (fun (v,t) -> (Local_var v, t)) vars in
          (* construct a block *)
          (* Block(Array_pattern var_pattern,  *)
          (*       [Type_decl([],[Call(List.hd var_exps, operation, *)
          (*                           List.tl var_exps, None),return_type]), *)
          (*        return_type]), *)
          Proc(Array_pattern var_pattern, 
                (Type_decl([],[Call(List.hd var_exps, operation,
                                    List.tl var_exps, None),return_type]),
                 return_type)),
          return_type
        | e -> invalid_arg (Printf.sprintf "unexpected block %s"
                              (string_of ~show_type:true e)) in

      fun env e -> match e with

       (* Arra.new(exp1,...,expN){ |var1,..,varN| [exp0,...,expM] }
          ==>
          Ikra_Zipper.new_alloc([exp1,...,expN], M){
            |ptr0,...,ptrM,size1,..,sizeN,lv1,...lvI| 
            M.initializeL(ptr0,...,ptrM,size1,..,sizeN,lv1,...lvI)
          }
       *)
       | Call((Constant(cname),t_rcv), mname, args, Some blk),
         Ikra_type.Tarray(rank,Ikra_type.Ttuple element_types)
         when cname = "PArray" || cname = "Array"
                    && mname = "new" || mname = "new_" ->
         let num_element_types = List.length element_types in
         let open Ast.Untyped_constructor in
         Some(Convert_kernel.Initialize,
              (dup num_element_types "ptr" tobject)@(dup rank "size" tint),
              "new_alloc", (Constant("Ikra_Zipper"),t_rcv),
              [array(args);literal(Integer_literal(num_element_types))],
              explicit_block blk)


       (* Array.new(exp1,..,expN){ |var1,..,varN| exp0 } ==> *)
       (* PArray.new(exp1,..,expN){ |ptr,size1,..,sizeN| *)
       (*    M.initializeL(ptr,size1,...,sizeN,lv1,..lvI) } *)
       (* where M and L are the module name and the number of generated *)
       (* function, lv1,..,lvI are the lexical variables in exp0. *)
       (* NOTE the class name can either be Array or PArray *)
       | Call((Constant(cname),t_rcv), mname, args, Some blk),
         Ikra_type.Tarray(rank,_)
         when cname = "PArray" || cname = "Array"
                    && mname = "new" || mname = "new_" ->
         Some(Convert_kernel.Initialize,
              ("ptr",tobject)::(dup rank "size" tint),
              mname, (Constant("PArray"),t_rcv), args, explicit_block blk)

       (* var0:array(N) |- var0.reduce(expZ){ |var1,var2| expB } ==> *)
       (* var0.reduce(expZ){ |ptr,size1,..,sizeN| *)
       (*   M.reduceL(ptr,size1,..,sizeN,lv1,...,lvI) } *)
       | Call((_,Ikra_type.Tarray(rank,_)) as rcv,
              method_name,
              ([_] as args),
              Some blk),_
         when (method_name = "reduce" || method_name = "inject") ->
         Some(Convert_kernel.Reduce,
              ("ptr",tobject)::("zero", tobject)::(dup rank "size" tint),
              "reduce", rcv, args, explicit_block blk)

       (* var0:array(N) |- var0.reduce{ |var1,var2| expB } ==> *)
       (* var0.unitless_reduce(){ |ptr,size1,..,sizeN| *)
       (*   M.reduceL(ptr,size1,..,sizeN,lv1,...,lvI) } *)
       | Call((_,Ikra_type.Tarray(rank,_)) as rcv, method_name, ([] as args),
              Some blk),_
         when (method_name = "reduce" || method_name = "inject") ->
         Some(Convert_kernel.UnitlessReduce,
              ("ptr",tobject)::(dup rank "size" tint),
              "reduce", rcv, args, explicit_block blk)

       (* ad-hoc support for neighbors *)
       (* var0:array(N) |- var0.map{ |varE,var1,...,varN| expB } ==> *)
       (* var0.map{ |ptr,size1,..,sizeN| *)
       (*   M.mapL(ptr,size1,..,sizeN,lv1,...,lvI) } *)
       | Call((_,(Ikra_type.Tarray(rank,Ikra_type.Tneighbor(_,_,t1)))) as rcv,
              "map", args, Some blk),t2  ->
         Some(Convert_kernel.Map,
              ("ptr_array", Ikra_type.Tarray(rank,t1))
              ::("ptr_pad", t1)
              ::("ptr2", t2)
              ::(dup rank "size" tint),
              "map", rcv, args, explicit_block blk)

       (* var0:array(N) |- var0.map{ |varE,var1,...,varN| expB } ==> *)
       (* var0.map{ |ptr,size1,..,sizeN| *)
       (*   M.mapL(ptr,size1,..,sizeN,lv1,...,lvI) } *)
       | Call((_,
               (Ikra_type.Tarray(rank,_) as t1))
                 as rcv,
              "map", [], Some blk),
         (Ikra_type.Tarray(_,Ikra_type.Ttuple(element_types)) as t2)  ->
         let open Ast.Untyped_constructor in
         Some(Convert_kernel.Map,
              ("ptr", t1)::("ptr2", t2)::(dup rank "size" tint),
              "map_zipped", rcv, [
                literal(Integer_literal(List.length element_types))
              ], explicit_block blk)

       (* var0:array(N) |- var0.map{ |varE,var1,...,varN| expB } ==> *)
       (* var0.map{ |ptr,size1,..,sizeN| *)
       (*   M.mapL(ptr,size1,..,sizeN,lv1,...,lvI) } *)
       | Call((_,(Ikra_type.Tarray(rank,_) as t1)) as rcv,
              "map", args, Some blk),t2  ->
         Some(Convert_kernel.Map,
              ("ptr", t1)::("ptr2", t2)::(dup rank "size" tint),
              "map", rcv, args, explicit_block blk)
       | _ -> None

let expand_tuples
    : (string * Ikra_type.t) list ->(string * Ikra_type.t) list =
  (* given a list of pairs of variable names and of their types,
     construct an identical list except that each array of tuples are
     replaced with the list of arrays, each of which represent
     elements of the tuples.

     expand_tuples [(v1,t1); ...; (vi,Tarray(n,Ttuple(s1,..,sM)); ...; (vN,tN)]
     = [(v1,t1); ...; (vi_t1,Tarray(n,s1));...(vi_tM,Tarray(n,sM));
     ...; (vN,tN)]
  *)
  let open Ikra_type in
  let rec elements = function
    | Ttuple typs -> List.concat (List.map elements typs)
    | t -> [t] in
  
  fun vars_types -> 
  List.flatten
    (List.map (function
      | v,Tarray(n,(Ttuple(element_types) as t)) ->
        Util.mapi (fun i et -> (v ^ "_t" ^ (string_of_int i), et))
          (elements t (* element_types *))
      | vt -> [vt])
       vars_types)

let replace_array_operations :
    string -> tast -> tast * (array_ops list) =
  (* replace_array_operations source returns a copy of the source AST
     in which array operations are replaced with the calls to the CUDA
     functions. *)
  let open Ast in
      let pass_lexical_vars : tast -> Ikra_type.t Env.env -> tast list =
        (* pass_lexical_vars exp = [Local_var v1;...]  where
           v1,... are the free variables in exp. The result is used as
           additional arguments to the kernel function. *)
        fun exp env ->
          List.map (fun v -> (Local_var v, Env.lookup env v))
            (List.sort String.compare
               (Ast.VarSet.elements (Ast.free_variables exp))) in

      let rec extend_formal_params
          : Ikra_type.t Env.env
        -> parameter_pattern list -> Ikra_type.t  Env.env =
        fun e ps ->
          List.fold_left (fun env -> (function
            | Var_pattern(v,t) -> Env.extend env [v,t]
            | Array_pattern pats -> extend_formal_params env pats))
            e ps in

      fun module_name typed_exp ->

        let array_ops = ref [] in

        let push : Convert_kernel.kind -> string -> tast list -> tast
          -> Ikra_type.t -> tast -> unit =
          (* (push name expression free_vars receiver)
             remembers the expression under the
             name for the later CUDA code generation *)
          fun kind name free_vars receiver return_type -> function
            (* | Block(params, [Type_decl(decls, exps),typ]),_ -> *)
            | Proc(params, (Type_decl(decls, exps),typ)),_ ->
              (* NOTE: the return type of the block is the return type
                 of its body; typeinf does not give a type to block
                 itself (for now).  *)
              let kern = {
                Convert_kernel.kind        = kind;
                Convert_kernel.name        = name;
                Convert_kernel.free_vars   = free_vars; 
                Convert_kernel.params      = params;    
                Convert_kernel.type_decls  = decls;     
                Convert_kernel.body        = exps;
                Convert_kernel.receiver    = receiver;
                Convert_kernel.body_type   = typ;
                Convert_kernel.top_decls   = typed_exp;
                Convert_kernel.return_type = return_type;
              } in
              array_ops := kern::!array_ops
            | e -> invalid_arg ("unexpected block" ^ (Ast.string_of e)) in
        


        let gensym : string -> string =
          (* gensym base constructs a unique name that begins with base,
             followed by a unique number.  The number starts from zero.
             This must be defined inside the function in order to reset
             the counter for every time of conversion.*)
          let counter = ref 0 in
          fun base ->
            let n = !counter in
            let _ = counter := n+1 in
            base ^ (string_of_int n) in

        (* body of the function *)
        let rec replace : Ikra_type.t Env.env -> tast -> tast =
          fun env -> function

            (* array operation like map, reduce, and PArray.new whose
               RETURN TYPES ARE ARRAY OF TUPLES *)
            (* | Call(_,_,args,_), *)
            (*   (Ikra_type.Tarray(_,Ikra_type.Ttuple _) as t) as e *)
            (*   when Util.Option.is_some(array_op env e) -> *)

            (*   let ftype, vars, meth, recv, (block,t2) *)
            (*     = Util.Option.get(array_op env e) in *)
            (*   let fname = gensym (Convert_kernel.string_of_kind ftype) in *)
            (*   let fvs = pass_lexical_vars (block,t2) env in *)
            (*   let _ = push ftype fname fvs recv t (block,t2) in *)

            (*   (\* a variable of type array of tules, which may be *)
            (*      contained in vars, should be expanded to a sequence *)
            (*      of variables, each of which corresponds to (an array *)
            (*      of) element in the tuple. *\) *)
              
            (*   (\* TODO: expand the array of tuples before pushing, then *)
            (*      we don't need to expand it at generating CUDA *)
            (*      code. *\) *)

            (*   let vars = expand_tuples vars in *)

            (*   let open Ast.Untyped_constructor in *)
            (*       (Call(((replace env) recv), *)
            (*             meth, List.map (replace env) args, *)
            (*             Some(Proc(Ast.Array_pattern *)
            (*                          (List.map (fun (v,_) -> *)
            (*                            var_pattern (v (\* TODO confirm if this is ok *\))) *)
            (*                         vars), *)
            (*                    call((constant(module_name)), *)
            (*                          fname, *)
            (*                          (List.map (fun (v,t) -> (Local_var v,t)) *)
            (*                             vars) @ fvs, *)
            (*                          None)), *)
            (*              t2)), *)
            (*    t) *)




            (* array operation like map, reduce, and PArray.new *)
            | Call(_,_,_,_),t as e
              when Util.Option.is_some(array_op env e) ->
              let ftype, vars, meth, recv, args, (block,t2)
                = Util.Option.get(array_op env e) in
              let fname = gensym (Convert_kernel.string_of_kind ftype) in
              let fvs = pass_lexical_vars (block,t2) env in
              let _ = push ftype fname fvs recv t (block,t2) in

              (* a variable of type array of tules, which may be
                 contained in vars, should be expanded to a sequence
                 of variables, each of which corresponds to (an array
                 of) element in the tuple. *)
              
              (* TODO: expand the array of tuples before pushing, then
                 we don't need to expand it at generating CUDA
                 code. *)

              let vars = expand_tuples vars in

              let open Ast.Untyped_constructor in
                  (Call(((replace env) recv),
                        meth, List.map (replace env) args,
                        Some(Proc(Ast.Array_pattern
                                     (List.map (fun (v,_) ->
                                       var_pattern (v (* TODO confirm if this is ok *)))
                                    vars),
                               call((constant(module_name)),
                                     fname,
                                     (List.map (fun (v,t) -> (Local_var v,t))
                                        vars) @ fvs,
                                     None)),
                         t2)),
               t)


            (* the rests are just applying the replacement to all
               sub-expressions *)

            (* method call *)
            | Call (receiver, method_name, operands, block),t ->
              Call((replace env) receiver,
                   method_name,
                   List.map (replace env) operands,
                   maybe (replace env) block),t
            (* literal  *)
            | Literal(literal),t -> Literal(literal),t
            (* assignment to a constant  *)
            | Constant_assign(name,ast) as e,t ->
              if Env.is_bound env name
              then Constant_assign(name,replace env ast),t
              else invalid_arg
                ("no type info: " ^ name ^ " in "^ (string_of (e,t))
                 ^ " under env of " ^ (String.concat "," (Env.bound_vars env) ))
            (* constant *)
            | Constant (name),t -> Constant (name),t
            (* assignment to a local variable  *)
            | Local_assign(name,ast) as e,t ->
              if Env.is_bound env name
              then Local_assign(name,replace env ast),t
              else invalid_arg
                ("no type info: " ^ name ^ " in "^ (string_of (e,t))
                 ^ " under env of " ^ (String.concat "," (Env.bound_vars env) ))
            (* local variable reference *)
            | Local_var (name),t ->  Local_var (name),t
            (* block structure*)
            (* | Block (variables, body),t -> *)
            (*   Block (variables, List.map (replace env) body),t *)
            | Block (body),t ->
              Block (List.map (replace env) body),t
            (* procedure as a parameter *)
            | Proc (parameters, body),t ->
              Proc (parameters, replace env body),t
            (* method definition *)
            | Method_def (method_name, variables, body),t ->
              let env = extend_formal_params Env.empty [variables] in
              Method_def (method_name, variables, replace env body),t
            (* yield *)
            | Yield (ast),t -> Yield (replace env ast),t
            (* array literal *)
            | Array (exps),t -> Array (List.map (replace env) exps),t
            (* while loop *)
            | While (cond, body),t ->
              While (replace env cond, replace env body),t
            (* conditional branch *)
            | If (cond, then_exp, else_exp),t ->
              If (replace env cond, replace env then_exp,
                  maybe (replace env) else_exp),t
            (* for loop  *)
            | For(var, range, body),t ->
              For(var, replace env range, replace env body),t

            (* return *)
            | Return(exp1),t -> Return(maybe (replace env) exp1),t

            (* type_decl *)
            | Type_decl (tdecls, body),t ->
              let exenv =  (Env.extend env tdecls) in
              Type_decl (tdecls,
                         List.map (replace exenv) body),t

        in let converted = replace Env.empty typed_exp in
           (converted, List.rev !array_ops)
         

(* NOTE: before the code generation, we need to replace a symbol block
   parameter with an explicit block.  E.g., e0.m(e1..en,&:n) will
   become e0.m(e1..en){ |x1..xm| x1.n(x2..xm) } where &:n has a type
   of m-ary procedures.  This replacement is only when the call is
   selected as an array operation.
*)

let generate_ruby_file :
    Convert_env.t -> tast -> Convert_kernel.kernel list =
  (* generate_ruby_file ruby_module_name filename typed_source_program writes the Ruby
     part of the given typed_source_program into the given filename,
     where operations on the parallel arrays are replaced with the
     calls to CUDA code. *)

  fun gen_env typed_source_program -> 
    let ochan = open_out (Convert_env.get_ruby_file_name gen_env) in
    let converted, kernels =
      replace_array_operations (Convert_env.get_module_name
                                  gen_env)
        typed_source_program in
    (* The following header loads the PArray library and the generated
       kernel functions.
       
       Line 1 is the standard shebang-line.

       Lines 2 and 3 load .ikra.conf in the current or home directory.

       Line 4 sets the backend name, either c or cuda, which changes
       the behavior of the parray library.

       Line 5 adds a path to the parray library to $LOAD_PATH.

       Line 6 loads the parray library.

       Line 7 loads the compiled kernel functions in the same
       directory as the one where the converted Ruby file is located.
       It uses an absolute path because the ruby process that executes
       the converted Ruby file may run under the directory where the
       original Ruby program is located (so that it can load other
       files from the current directory).

    *)
    Printf.fprintf ochan "\
#!/bin/env ruby
['.',ENV['HOME']].find do |d| p=File.expand_path('.ikra.conf',d)
File.exists?(p) && load(p) end
$IKRA_BACKEND='%s'
$LOAD_PATH.push(File.expand_path('%s',IKRA_HOME))
require('parray')
require(File.expand_path('%s',File.dirname(__FILE__)))
#END_OF_IKRA_HEADER
"
      Convert_kernel.backend_name
      (Config.parray_dir_relative ())
      (Convert_env.get_require_name gen_env);
    output_string ochan (Ast.string_of ~hide_type_decl:true converted);
    close_out ochan;
    kernels

let generate_cuda_file
    : ?generate_only:bool -> Convert_env.t  -> Convert_kernel.kernel list
  -> unit =
  (* (generate_cuda_file gen_env kernel_infos) writes the CUDA part of
     the given kernel_infos into the CUDA file whose name is specified
     by the gen_env.  The file contains the wrapper, iterator and
     kernel functions for each parallel array operation. *)

  fun ?(generate_only=false) gen_env kernels ->
    let ochan = open_out (Convert_env.get_cuda_file_name gen_env) in
    output_string ochan (Convert_kernel.cuda_code_of
                           (Convert_env.get_require_name gen_env)
                           kernels);
    close_out ochan;
    Convert_kernel.generate_makefile ~generate_only gen_env

let target_file_names :
    ?cleanup:bool -> string option -> string -> (Convert_env.t -> 'a) -> 'a =
  (* (target_file_names target_dir source_file_name proc) creates a
     set of names for the code to be generated, and runs proc with the
     set. Generated files are stored in the target_dir if it is
     specified, otherwise, they are in a newly created temporary
     directory.

     Example: (target_file_names "reduction.rb" proc) ->
     proc { ruby_require_name = "reduction",
     ruby_module_name = "Ikra_Kernels__reduction",
     ruby_file_name = "/tmp/reductio12345/reduction.ikra.rb",
     cuda_file_name = "/tmp/reductio12345/reduction.ikra.cu" }
  *)

  fun ?(cleanup=true) target_dir source_file_name proc ->
    let open Str in
        let basename =
          (global_replace (regexp "\\.rb$") ""
             (Filename.basename source_file_name)) in
        let require_name =
          global_substitute (regexp "[^A-Za-z0-9_]") (fun _ -> "_")
            basename in
        let dir = match target_dir with
          | Some d -> d
          | None   -> Util.make_temp_dir "ikragen" in
        let generator suffix =
          (Filename.concat dir 
             (String.concat "." [basename; suffix])) in
        let names = 
          let open Convert_env in
              {
                require_name = require_name;
                module_name = "Ikra_Kernels__" ^ require_name;
                ruby_file_name = generator target_ruby_suffix;
                cuda_file_name = generator target_cuda_suffix;
                build_script_name = "build.sh";
              } in
          Util.with_finally (fun () -> proc names)
            (fun () ->
              if cleanup
              then ignore (Sys.command ("rm -r " ^ dir))
              else ())

let generate_and_compile
    : ?generate_only:bool -> Convert_env.t -> tast -> unit =
  (* (generate_and_compile module_name ruby_module_name ruby_filename
     cuda_filename typed_source_program) compiles the given
     typed_source_program into Ruby and CUDA programs, which are
     stored in the given ruby_filename and cuda_filename.  The
     generated CUDA module can be accessed through the given
     module_name from Ruby. *)

  fun ?(generate_only=false) gen_env typed_source_program ->
    let kernels = generate_ruby_file gen_env typed_source_program in
    generate_cuda_file ~generate_only gen_env kernels

let run_converted
    : ?generate_only:bool -> ?redirect:bool -> Convert_env.t -> string list
  -> string =
  fun ?(generate_only=false) ?(redirect=true) gen_env app_args ->
    let runner ruby_file_name =
      if generate_only then
        (let args = if app_args=[]
          then Printf.eprintf "ignoring application arguments: %s\n"
            (String.concat " " app_args)
          else ();
           ["$@"] in
         let ochan = open_out_gen [Open_text;Open_append;Open_wronly] 0o755
           (Filename.concat
              (Filename.dirname ruby_file_name)
              (Convert_env.get_build_script_name gen_env)) in
         output_string ochan "if [ \"$1\" != \"--no-exec\" ]; then\n";
         output_string ochan (String.concat " " 
                                ("ruby"::"-I" ::"`dirname \"$0\"`"
                                 ::(Filename.basename ruby_file_name)::args));
         output_string ochan "\n";
         output_string ochan "fi\n";
         close_out ochan;
         "written.")
      else Util.run_ruby ~redirect (ruby_file_name::app_args) in
    let ruby_file_name = Convert_env.get_ruby_file_name gen_env in
    (* run the converted Ruby file.  Note that the current working
       directory is not changed from the one where the Ikra compiler
       is invoked, while the converted script (as specified by an
       absolute path in ruby_file_name) may be created in a different
       (possibily temporary) directory. *)
    runner ruby_file_name

end
