open Ast
open Ikra_type

(* The RubyParser returns a parsed program in an S-expression, which
   is represented in the following style:

   s(:call, nil, :puts, s(:arglist, s(:call, s(:lit, 0), :+, s(:arglist,
   s(:lit, 3)))))

   The syntax is therefore like this:

   Sexp ::= "s" "(" Sexp_list ")"
        ::= "nil"
	::= ":" Symbol
	::= Number

   Sexp_list ::= (empty) 
	     ::= Sexp
	     ::= Sexp "," Sexp_list

*)

let path_to_parser _ =
  Filename.concat (Config.ikra_root_directory()) "src/parser.rb"

type sexp =
  | Sexp_symbol of string
  | Sexp_nil
  | Sexp_list of sexp list
  | Sexp_int of int
  | Sexp_float of float
  | Sexp_str of string
  | Sexp_bool of bool
  | Sexp_range of int * int

let rec string_of_sexp : sexp -> string =

  function
    | Sexp_symbol s -> ":" ^ s
    | Sexp_nil -> "nil"
    | Sexp_int n -> string_of_int n
    | Sexp_float n -> string_of_float n
    | Sexp_str s -> "\"" ^ s ^ "\""
    | Sexp_list sexps -> "s(" ^ (String.concat ", " 
				   (List.map string_of_sexp sexps)) ^ ")"
    | Sexp_bool b -> string_of_bool b
    | Sexp_range(x,y) -> (string_of_int x)^".."^(string_of_int y)

let symbol_of_sexp = function
  | Sexp_symbol s -> s
  | sexp ->
    raise (Invalid_argument ("not a Sexp_symbol:"^(string_of_sexp sexp)))

let ruby_parse_file : string -> string =
  (* parse a Ruby program in the given filename by using Ruby parser,
     and returns an SEXP as a string *)

  fun filename -> Util.run_ruby [path_to_parser (); filename]

let literal_range_pattern =
  (* literal dot2 notation like 0..9; For some reasons, RubyParser
     parses an e1..e2 expression as a literal when e1 and e2 are
     integers.  *)
  "\\([0-9]+\\)\\.\\.\\([0-9]+\\)"

let literal_range_regexp = Str.regexp literal_range_pattern

(* TODO: check the lexical analyzer of RubyParser so as to cover all
   tokens in the language. *)

let tokenizing_regexp =
  Str.regexp 
    ("[ \t]*\\(" ^
	(String.concat "\\|"
	   ["\\$[A-Za-z_0-9]+";
            "\\$?[A-Za-z_][A-Za-z_0-9]*";
	    "\\+@?";
            "-@";
            "\\*\\*";
	    "[*/%():,^&|~]";
	    "\"[^\"]*\"";
	    literal_range_pattern;
	    "-?[0-9]+\\(\\.[0-9]+\\)?\\([Ee][-+][0-9]+\\)?";
            "-";
	    "<<"; ">>";
	    "[<>=]=?";
	    "\\[\\]";			(* array reference operator *)
	   ]) ^
	"\\)[ \t]*")

let rec tokenize_sexp : string -> string list =
  (* decompose the given sexp into a list of tokens *)
  
  function
    | "" -> []
    | sexp ->
      let r = tokenizing_regexp in
	 if Str.string_match r sexp 0
	 then let s = Str.matched_group 1 sexp in
	      s::(tokenize_sexp (Str.string_after sexp (Str.match_end ())))
	 else raise (Invalid_argument sexp)

(* TODO: use regular expressions *)
let parse_literal : string -> sexp =
  (* parse literal token as an sexp *)

  fun token -> match String.get token 0, token with
    (* " CHAR* " *)
    | '"',_ -> Sexp_str (String.sub token 1 (String.length token - 2))
    (* DIGIT+ .. DIGIT+ *)
    | _ when Str.string_match literal_range_regexp token 0 ->
      let from_num = int_of_string(Str.matched_group 1 token) in
      let to_num   = int_of_string(Str.matched_group 2 token) in
      Sexp_range(from_num, to_num)
    (* DIGIT+ (. DIGIT+)? *)
    | ('-'|'0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9'),_ ->
      (try let _ = String.index token '.' in
	   Sexp_float (float_of_string token)
       with Not_found ->
	 Sexp_int (int_of_string token))
    (* true | false *)
    | _,("true"|"false") -> Sexp_bool (bool_of_string token)
    | _ -> raise (Invalid_argument ("parse_literal" ^ token))


let rec parse_sexp : string list -> sexp * string list =
  (* parse tokens as an sexp and the remaining tokens *)

  fun tokens -> match tokens with
    | "nil"::rest -> Sexp_nil, rest
    | ":"::word::rest -> Sexp_symbol word, rest
    | "s"::"("::rest ->
      (match parse_sexp_list [] rest with
	| (sexps, ")"::rest) -> Sexp_list sexps, rest
	| _ -> raise (Invalid_argument (String.concat ";" tokens)))
    | word::rest ->
      (parse_literal word, rest)
    | _ -> raise (Invalid_argument (String.concat ";" tokens))

and parse_sexp_list :
    sexp list -> string list -> sexp list * string list =
  (* parse tokens as a list of sexps and the remaining tokens *)

  fun sexps tokens -> match tokens with
    | ")"::rest -> sexps, tokens
    | _ ->
      (match parse_sexp tokens with
	| sexp, ")"::rest -> sexps@[sexp], ")"::rest
	| sexp, ","::rest -> parse_sexp_list (sexps@[sexp]) rest
	| _ -> raise (Invalid_argument (String.concat ";" tokens)))


let parse_sexp_string : string -> sexp =
  (* parse a string representation of sexp into an sexp *)

  fun sexp ->
    let sexp, _ = parse_sexp (tokenize_sexp sexp) in
    sexp

let add_block : (tast (* Call *) * tast (* Block *)) -> tast (* Call *) =
  (* add the given block to the given call ast *)

  function
    | (Call(receiver, method_name, operands, None),t1),
	((Proc _, t2)as block) ->
      (Call(receiver, method_name, operands, Some (block)), t1)
    | (call),(block) ->
      raise (Invalid_argument 
	       ("unexpected nodes for add_block: "^
		   (Ast.string_of call) ^ " and " ^
		   (Ast.string_of block)))


let rec ast_of_sexp_maybe : sexp -> tast option =
  (* parse the given sexp if it is not nil; otherwise, it return
     None *)

  function (* sexp -> match sexp with *)
    | Sexp_nil -> None
    | sexp     -> Some(ast_of_sexp sexp)

and ast_of_sexp : sexp -> tast =
  (* convert sexp into ast *)
  let open Ast.Untyped_constructor in
  fun sexp ->
    let fail _ = raise (Invalid_argument (string_of_sexp sexp)) in
    match sexp with

      (* Exp . Identifier( Exp* ) --- method call with a receiver *)
      | Sexp_list[Sexp_symbol "call";
		  receiver;
		  Sexp_symbol mname;
		  Sexp_list(Sexp_symbol "arglist"::operands)] ->
        let receiver' = match receiver with
          |Sexp_nil -> implicit_self
          |sexp -> (ast_of_sexp sexp) in
        let operands', block = parameter_list_of_sexp operands in
        call(receiver', mname, operands', block)

      (* Exp . Identifier ( Exp* ) { |Identifier*| Exp* } *)
      | Sexp_list[Sexp_symbol "iter";
		  call_sexp;
		  block_parameters_sexp;
		  block_sexp] ->
	add_block (ast_of_sexp call_sexp,
		   proc_of_sexp block_parameters_sexp block_sexp)

      (* Number -- intger literal *)
      | Sexp_list[Sexp_symbol "lit"; Sexp_int n] -> 
	literal(Ast.Integer_literal n)

      (* Number -- intger literal *)
      | Sexp_list[Sexp_symbol "lit"; Sexp_float n] -> 
	literal(Ast.Float_literal n)

      (* Symbol -- symbol literal *)
      | Sexp_list[Sexp_symbol "lit"; Sexp_symbol s] -> 
	literal(Ast.Symbol_literal s)

      (* Special variable __FILE__ *)
(*
Strangely, the Ruby parser cannot properly read the variable
__FILE__.
$ echo "__FILE__" | ruby parser.rb
s(:str, "(string)")
*)
      | Sexp_list[Sexp_symbol "str"; Sexp_str "(string)"] -> 
	constant "__FILE__"

      (* String -- string literal *)
      | Sexp_list[Sexp_symbol "str"; Sexp_str s] -> 
	literal(Ast.String_literal s)

      (* Constant *)
      (* TODO: we'd better to make anything that begins with uppcase
         letters a constant.  We currently parse selected names as
         constants because of user defined constants.  However, since
         there are almost unlimited number of constants in Ruby, it
         would be safe to first parse every uppercase-beginnig-name as
         constant, and then switch to a variable after we found an
         assignment. *)
      | Sexp_list[Sexp_symbol "const"; Sexp_symbol name] ->
        (* if false then *)
        (* let constants = ["Array";"Math";"Integer";"Float";"Array";"Object";"Neighbor";"Tuple"; "PArray"; "Time"; "ARGV"; ]in (\* NOTE: make it general *\) *)
        (* if List.mem name constants *)
        (* then constant name *)
        (* else local_var name                  (\* NOTE: is this safe? *\) *)
        (* else *) constant name

      | Sexp_list[Sexp_symbol name]
        when List.mem name ["true"; "false"] ->
        constant name

      | Sexp_list[Sexp_symbol "colon2"; modul; Sexp_symbol cname]
          when is_constant(ast_of_sexp modul) ->
        (match ast_of_sexp modul with
          |Constant mname, _ ->
            constant (mname ^ "::" ^ cname)
          |_ -> invalid_arg "never happens")

      (* Id = Exp -- assignment to local var *)

      | Sexp_list[Sexp_symbol "lasgn"; Sexp_symbol name; exp] ->
	local_assign(name, ast_of_sexp exp)

      (* Id = Exp --- assignment to a constant *)
      (* NOTE: we use local_assign for constants *)
      | Sexp_list[Sexp_symbol "cdecl"; Sexp_symbol name; exp] ->
	constant_assign(name, ast_of_sexp exp)

      (* Id  -- reference to local var *)

      | Sexp_list[Sexp_symbol "lvar"; Sexp_symbol name] ->
	local_var(name)

      (* Id -- reference to global var ; global variables are treated
         as local variables *)
      | Sexp_list[Sexp_symbol "gvar"; Sexp_symbol name] ->
	constant name

      (* Id -- self is a pseudo variable, but treated as a local
	 variable in ASTs *)
      | Sexp_list[Sexp_symbol "self"] ->
	local_var "self"

      (* Exp; Exp; ... ; Exp -- sequence *)

      | Sexp_list(Sexp_symbol "block"::sexps) ->
	block(List.map ast_of_sexp sexps)

      (* def Id ( Id, Id, ... Id ); Exp; ... Exp; end -- method def. *)

      | Sexp_list[Sexp_symbol "defn";
		  Sexp_symbol method_name;
		  Sexp_list(Sexp_symbol "args" :: args );
		  Sexp_list[Sexp_symbol "scope"; body]] ->
	method_def(
	  method_name,
	  array_pattern 
	    (List.map (fun s -> var_pattern (symbol_of_sexp s)) args),
	  ast_of_sexp body)

      (* yield Exp *)

      | Sexp_list[Sexp_symbol "yield"; sexp] ->
	yield(ast_of_sexp sexp)

      (* [ Exp, Exp, ..., Exp ] -- array literal *)
	  
      | Sexp_list(Sexp_symbol "array"::sexps) ->
	array(List.map ast_of_sexp sexps)

      (* while Exp; Exp ... Exp; end *)

      | Sexp_list[Sexp_symbol "while"; cond_sexp; body_sexp;
		  (* RubyParser generates an sexp for a while
		     statement with condition, body, and a constant true,
		     whose meaning is not documented. *)
		  Sexp_bool true] ->
	while_(ast_of_sexp cond_sexp,
		  ast_of_sexp body_sexp)

      (* if Exp then Exp else Exp end *)
      | Sexp_list[Sexp_symbol "if"; cond_sexp; then_sexp; else_sexp] ->
	if_(ast_of_sexp cond_sexp,
	       ast_of_sexp then_sexp,
	       ast_of_sexp_maybe else_sexp)

      (* for Id in Exp do Exp end -- we only support the case when a
	 loop variable is a local variable, while Ruby seems to
	 support other L-values like a field of an object, an element
	 of an array as a loop variable. *)

      | Sexp_list[Sexp_symbol "for"; range_sexp;
		  (* if a loop variable is a local variable, this
		     field will become an "lasgn" node. *)
		  Sexp_list[Sexp_symbol "lasgn"; Sexp_symbol var_name];
		  body_sexp] ->
	for_(var_name, ast_of_sexp range_sexp, 
		ast_of_sexp body_sexp)

      (* return exp / return *)
      | Sexp_list[Sexp_symbol "return"; exp1] -> return (ast_of_sexp exp1)
      | Sexp_list[Sexp_symbol "return";] -> return0 ()

      (* TODO confirm if self insertion is correct *)
      (* Exp .. Exp -- a range expression  *)
      | Sexp_list[Sexp_symbol "dot2"; from_sexp; to_sexp] ->
	call(implicit_self, Ast.pseudo_name_of "..",
		 [ast_of_sexp from_sexp;
		  ast_of_sexp to_sexp], None)

      (* TODO confirm if self insertion is correct *)
      (* Num .. Num -- a (literal) range expression  *)
      | Sexp_list[Sexp_symbol "lit"; Sexp_range(f,t)] ->
	call(implicit_self, Ast.pseudo_name_of "..",
		 [literal(Ast.Integer_literal f);
		  literal(Ast.Integer_literal t)], None)

      | Sexp_list[Sexp_symbol "not"; exp] ->
	call(implicit_self, Ast.pseudo_name_of "!", [ast_of_sexp exp], None)
	(* call(ast_of_sexp exp, Ast.pseudo_name_of "!", [], None) *)

      | Sexp_list[Sexp_symbol "and"; exp1; exp2] ->
	call((ast_of_sexp exp1), "&&", [ast_of_sexp exp2;], None)
        (* call(implicit_self, "&&", *)
        (*      List.map ast_of_sexp [exp1;exp2;], None) *)

      | Sexp_list[Sexp_symbol "or"; exp1; exp2] ->
	call((ast_of_sexp exp1), "||", [ast_of_sexp exp2;], None)
        (* call(implicit_self, "||", *)
        (*      List.map ast_of_sexp [exp1;exp2;], None) *)

      | _ -> fail ()

and block_parameters_of_sexp : sexp -> parameter_pattern =
  (* parse block parameters *)
    let open Ast.Untyped_constructor in
  function
    | Sexp_nil -> empty_pattern
    | Sexp_list [Sexp_symbol "masgn";
		 Sexp_list((Sexp_symbol "array")::patterns)] ->
      array_pattern (List.map block_parameters_of_sexp patterns)
    | Sexp_list[Sexp_symbol "lasgn"; Sexp_symbol var_name] ->
      var_pattern (var_name)
    | _ -> raise (Invalid_argument "block_parameters")

and proc_of_sexp : sexp -> sexp -> tast =
  (* parse porcedure parameter expression *)
    let open Ast.Untyped_constructor in
  fun parameters_sexp body_sexp ->
    let params = block_parameters_of_sexp parameters_sexp in
    proc(params, ast_of_sexp body_sexp)

and parameter_list_of_sexp : sexp list -> tast list * tast option =
  (* to parse a list of parameters.  When the list includes a
     block parameter, it appears in the second element of the
     return value. *)
  let rec loop = fun args -> function
    | [Sexp_list[Sexp_symbol "block_pass"; blk ]]
      -> List.rev args, Some (ast_of_sexp blk)
    | []
      -> List.rev args, None
    | hd::tl ->
      loop (ast_of_sexp hd::args) tl in
  fun sexps -> loop [] sexps

let parse_file : string -> tast =
  (* to build an AST of the program in the given filename *)

  fun filename ->
    let sexp_str = ruby_parse_file filename in
    let sexp     = parse_sexp_string sexp_str in
    ast_of_sexp sexp
    
