(* Tarjan's strongly connected components algorithm *)

type 'a node = { element : 'a; 
		 next : 'a node list ref;
		 index : int ref;
		 lowlink : int ref; }
type 'a graph = 'a node list

type 'a scc = 'a node list
type 'a tarjan_result = 'a scc list 

type 'a state_t = 'a tarjan_result * int * 'a node list

let make_node : 'a -> 'a node =
  fun element -> { element = element;
		   next = ref [];
		   index = ref (-1);
		   lowlink = ref (-1);}
    
let connect : 'a node -> 'a node -> unit =
  fun { next = next } n2 ->
    next := n2 :: !next

let string_of : 'a node -> string =
  fun { element=element; index=index; lowlink=lowlink } ->
    Printf.sprintf "%d(%d,%d)" element !index !lowlink

let string_of_scc : 'a node list -> string =
  fun nodes ->
    "[" ^ (String.concat ";" (List.map string_of nodes)) ^ "]"

let string_of_sccs : 'a node list list -> string =
  fun sccs ->
    String.concat ";" (List.map string_of_scc sccs)

let visited_p : 'a node -> bool =
  fun {index=iref} -> !iref >= 0

let rec strongconnect : 'a state_t -> 'a node -> 'a state_t =
  (* given a SCCs already found, the maximum index number used so far,
     a stack of visited nodes, and a starting node v, find SCCs
     containing nodes reachable from v*)
  
  fun (sccs,index,stack) v ->
    (* set the depth index of v to the smallest unused index *)
    v.index   := index;
    v.lowlink := index;
    (* visit neighbours of v *)
    let state = traverse v (sccs,index+1,v::stack) in
    (* if v is a root node, construct SCC from the stack *)
    if !(v.lowlink) = !(v.index) 
    then build_scc v [] state
    else state
and traverse v s =
  List.fold_left (fun ((_,_,stack) as state) w ->
    if (visited_p w) && List.mem w stack then
      (* successor w is in stack and hence in the current SCC *)
      (v.lowlink := min !(v.lowlink) !(w.index); state) 
    else if (visited_p w) then 
      state (* do nothing *)
    else (* successor w has not yet been visited; recurse on it *)
      let state' = strongconnect state w in
      v.lowlink := min !(v.lowlink) !(w.lowlink);
      state') s !(v.next) 
and build_scc v scc = function 
  | (sccs,index,w::stack) when v.element=w.element ->
    (* NB: the above comparison must be done by using elements, as
       comparison of nodes can result in an infinite loop. *)
    ((w::scc)::sccs, index, stack)
  | (sccs,index,w::stack) when v.element!=w.element ->
    build_scc v (w::scc) (sccs,index,stack)
  | _ -> invalid_arg "no root is found"


let tarjan : 'a graph -> ('a tarjan_result * int * 'a node list) =
  fun nodes ->
    List.fold_left (fun state node ->
      if (visited_p node)
      then state
      else strongconnect state node)
      ([], 0, []) nodes

let sccs : 'a list -> ('a * 'a) list -> 'a list list =
     (* to construct a list of strongly connected componets for a given
	list of elements, and a list of a pair of elements, *)
 fun elements relations ->
   let elements_to_nodes =
     List.map (fun e -> (e, make_node e)) elements in
   let get_node element = List.assoc element elements_to_nodes in
   let _ = List.map (fun (e1,e2) ->
     connect (get_node e1) (get_node e2)) relations in
   let (sccs,_,_) = tarjan (List.map (fun (_,n) -> n) elements_to_nodes) in
   List.map (List.map (fun {element=e} -> e)) sccs

       
