open OUnit2

module Convert = Convert.M(Convert_kernel)

let test_target_file_directories _ =
  let assure_no_makefile file =
    let d = Filename.dirname file in
    let m = Filename.concat d "Makefile" in
    assert_bool
      ("There should be no Makefile in the same directory with " ^
	  file)
      (not(Sys.file_exists m)) in
  Convert.target_file_names None "foorbar" (fun gen_env -> 
    (assure_no_makefile
       (Convert_env.get_ruby_file_name gen_env);
     assure_no_makefile
       (Convert_env.get_cuda_file_name gen_env)))

let test_make_temp_dir _ =
  let d = Util.make_temp_dir "ikragen" in
  (assert_bool ("should be a dircetory:" ^ d) (Sys.is_directory d);
   Unix.rmdir d)

let test_generate_ruby_file _ =
  Convert.target_file_names None "foobar" (fun gen_env ->
    let ruby_file = Convert_env.get_ruby_file_name gen_env in
    let _ = Convert.generate_ruby_file gen_env
      (Ast.Literal(Ast.Integer_literal 0), Ikra_type.Tint) in
    assert_bool ("should exist a generated file: "^ ruby_file)
      (Sys.file_exists ruby_file);
    assert_bool ("should have some text in:" ^ ruby_file)
      ((Unix.stat ruby_file).Unix.st_size > 0);
    let _ = Parser.ruby_parse_file ruby_file in (* check parsable *)
    ())


let run_converted : string -> string -> string =
  fun kernel_require_name converted_ruby_file ->
    Util.run_ruby [
      "-I"; (Config.parray_dir ());
      "-I"; (Filename.dirname converted_ruby_file);
      "-rparray"; "-r"; kernel_require_name;
      converted_ruby_file ]

let make_test_convert (source, converted) =
  ("covert_" ^ source)>::
    fun _ ->
      Convert.target_file_names None "test_convert" (fun gen_env ->
	let ruby_file = Convert_env.get_ruby_file_name gen_env in
	let _ =
	  Convert.generate_ruby_file gen_env
	    (Ast.recognize
	       (Typeinf.recognize_type_decls
		  (Test_util.parse_ruby_string source))) in
        let actual = try Parser.parse_file ruby_file
          with Invalid_argument s ->
            invalid_arg (Printf.sprintf "parse error %s while parsing\n%s\n"
			   s (Util.read_lines ruby_file)) in
	let expected = Test_util.parse_ruby_string converted in
	assert_equal ~printer:Ast.string_of expected actual)

let testdata_convert = [

  (* the body of a PArray call will be converted to a call on a kernel
     method. *)
  "_t(PArray.new(10){ |x| _typedecl(Integer, :x); x },[Array, 1, Integer])",
  "PArray.new(10){ |ptr,size0| Ikra_Kernels__test_convert.initialize0(ptr,size0) }";

  (* the kernel method call in the converted code should have
     lexical variables in the body as parameters. *)
  "_typedecl(Integer, :l); l=3; \
   _t(PArray.new(10){ |x| _typedecl(Integer, :x); x+l },[Array, 1, Integer])",
  "l=3; \
   PArray.new(10){ |ptr,size0| Ikra_Kernels__test_convert.initialize0(ptr,size0,l) }";

  (* the lexical variables sholud be sorted in the alphabetical order. *)
  "_typedecl(Integer, :l, :a, :b); b=4; l=3; a=9;\
   _t(PArray.new(10){ |x| _typedecl(Integer, :x); a+x+l+b },[Array, 1, Integer])",
  "b=4; l=3; a=9;\
   PArray.new(10){ |ptr,size0| Ikra_Kernels__test_convert.initialize0(ptr,size0,a,b,l) }";

  (* reduce calls on an array. *)
  "_typedecl([Array,1,Integer], :a); a=0; \
   _t(a,[Array,1,Integer]).reduce(0){ |x,y| _typedecl(Integer, :x, :y); x+y }",
  "a=0; \
   a.reduce(0){ |ptr,zero,size0| Ikra_Kernels__test_convert.reduce0(ptr,zero,size0)}";

  (* inject is synonym of reduce *)
  "_typedecl([Array,1,Integer], :a); a=0; \
   _t(a,[Array,1,Integer]).inject(0){ |x,y| _typedecl(Integer, :x, :y); x+y }",
  "a=0; \
   a.reduce(0){ |ptr,zero,size0| Ikra_Kernels__test_convert.reduce0(ptr,zero,size0)}";

  (* map on an array *)
  "_typedecl([Array,1,Integer], :a); a=0; \
   _t(a,[Array,1,Integer]).map{ |x| _typedecl(Integer, :x);x*x }",
  "a=0; \
   a.map{ |ptr,ptr2,size0| Ikra_Kernels__test_convert.map0(ptr,ptr2,size0)}";

  (* combination of new and map. *)
  "_typedecl([Array,1,Integer], :a, Integer, :l, :m);\
   l = m = 1;\
   a=_t(PArray.new(10){ |x| _typedecl(Integer, :x); x+l },[Array, 1, Integer]);\
   puts _t(a,[Array,1,Integer]).reduce(1){ |x,y|_typedecl(Integer, :x, :y); x*y*m }",
  "l = m = 1;\
   a=PArray.new(10){ |ptr,size0| \
                    Ikra_Kernels__test_convert.initialize0(ptr,size0,l) };\
   puts a.reduce(1){ |ptr,zero,size0| \
                    Ikra_Kernels__test_convert.reduce1(ptr,zero,size0,m) }";


  "_typedecl([Array, 1, Integer],:a,:b);\
   a=_t(PArray.new(10){ |x| _typedecl(Integer, :x); x },[Array, 1, Integer]);\
   b=_t(a,[Array,1,Integer]).map{ |x| _typedecl(Integer, :x); x+1 };\
   puts _t(b,[Array,1,Integer]).reduce(0){ |x,y| _typedecl(Integer, :x, :y); x+y }",
  "a=PArray.new(10){ |ptr,size0| \
  			Ikra_Kernels__test_convert.initialize0(ptr,size0) };\
   b=a.map{ |ptr,ptr2,size0| \
  			Ikra_Kernels__test_convert.map1(ptr,ptr2,size0) };\
   puts b.reduce(0){ |ptr,zero,size0| \
  			Ikra_Kernels__test_convert.reduce2(ptr,zero,size0) }";

  "_typedecl([Array, 1, Integer],:a);\
   a=_t(PArray.new(10){ |x| _typedecl(Integer, :x); x },[Array, 1, Integer]);\
   puts _t(_t(a,[Array,1,Integer]).map{ |x| _typedecl(Integer, :x); x+2 },[Array,1,Integer]).\
          reduce(0){ |x,y| _typedecl(Integer, :x, :y); x+y }",
  "a=PArray.new(10){ |ptr,size0| \
  			Ikra_Kernels__test_convert.initialize0(ptr,size0) };\
   puts a.map{ |ptr,ptr2,size0| \
  			Ikra_Kernels__test_convert.map2(ptr,ptr2,size0) }.\
          reduce(0){ |ptr,zero,size0| \
  			Ikra_Kernels__test_convert.reduce1(ptr,zero,size0) }";

  (* floating point number *)

  "_t(PArray.new(10){ |x| _typedecl(Integer, :x); x*1.5 },[Array,1,Float])",
  "PArray.new(10){ |ptr,size0| Ikra_Kernels__test_convert.initialize0(ptr,size0) }";

  (* method call *)
  (* ? *)


  (* zip support *)
  "_typedecl([Array,1,[Tuple,Integer,Integer]],:a);\
   a=0;
   _t(_t(a,[Array,1,[Tuple,Integer,Integer]])\
      .map{ |x,y| _typedecl(Integer, :x, :y, :_RETURN); x+y },\
      [Array,1,Integer]);",
  "a=0;
   a.map{ |ptr_t0,ptr_t1,ptr2,size0| Ikra_Kernels__test_convert.map0(ptr_t0,ptr_t1,ptr2,size0) }";

  (* nested zip *)
  "_typedecl([Array,1,[Tuple,[Tuple,Integer,Integer],Integer]],:a);\
   a=0;
   _t(_t(a,[Array,1,[Tuple,[Tuple,Integer,Integer],Integer]])\
      .map{ |(x,y),z| _typedecl(Integer, :x, :y, :z, :_RETURN); x+y+z },\
      [Array,1,Integer]);",
  "a=0;
   a.map{ |ptr_t0,ptr_t1,ptr_t2,ptr2,size0| Ikra_Kernels__test_convert.map0(ptr_t0,ptr_t1,ptr_t2,ptr2,size0) }";

  (* neighbor support *)
  "_typedecl([Array,1,[Neighbor,3,1,Integer]],:a);\
   a=0;
   _t(_t(a,[Array,1,[Neighbor,3,1,Integer]])\
      .map{ |m|  _typedecl([Neighbor,3,1,Integer], :m, Integer, :_RETURN); 
                m[-1]+m[0]+m[1] },\
      [Array,1,Integer]);",
  "a=0;
   a.map{ |ptr_array,ptr_pad,ptr2,size0| Ikra_Kernels__test_convert.map0(ptr_array,ptr_pad,ptr2,size0) }";


  (* tuple support *)
  "_t(PArray.new(10){ |x| _typedecl(Integer, :x); [x,x*1.5] },[Array,1,[Tuple,Integer,Float]])",
  "Ikra_Zipper.new_alloc([10],2){ |ptr0,ptr1,size0| Ikra_Kernels__test_convert.initialize0(ptr0,ptr1,size0) }";

  (* tuple support *)
  "_t(_t(a,[Array,1,Integer]).map{ |x| _typedecl(Integer, :x); [x,x*1.5] },[Array,1,[Tuple,Integer,Float]])",
  "a().map_zipped(2){ | ptr,ptr2_t0,ptr2_t1,size0 | Ikra_Kernels__test_convert.map0(ptr,ptr2_t0,ptr2_t1,size0)}";

]

let test_convert_exps = List.map make_test_convert testdata_convert
  

let suite =
  "suite">:::[
    "target_file_directories">::test_target_file_directories;
    "make_temp_dir"	     >::test_make_temp_dir;
    "generate_ruby_file"     >::test_generate_ruby_file;
  ]
    @ test_convert_exps			

let _ = Config.initialize None

let () =
  run_test_tt_main suite
