(* The test file for C++ backend. *)
module Convert = Convert.M(Convert_kernel)

open OUnit2

let c_special_symbols : string list =
  (* a list of special symbols in C.  The C language allows to insert
     any number of spaces around them.  It should be in the descending
     order of the lengths. *)
    [
      "..."; ">>="; "<<="; "++"; "+="; "%="; "&&"; "&="; "*="; "/="; "!="; "--";
      "->"; "-="; "||"; "|="; ">>"; ">="; "=="; "<="; "<<"; "~="; "^="; "+";
      "%"; "&"; "*"; "}"; "{"; "]"; "["; ")"; "("; "."; "/"; "?"; "!"; ":"; ";";
      ","; "-"; "|"; ">"; "="; "<"; "~"; "^";
    ]

let c_delimiter : Str.regexp =
  Str.regexp
    ("[ \t\n]*\\(" ^
	(String.concat "\\|"
	   (List.map
	      (fun sym -> Str.quote sym)
	      (List.sort (fun x y -> (String.length y) - (String.length x))
		 c_special_symbols))) ^
	"\\|[ \t\n]\\)[ \t\n]*")

let quote : string -> string =
  (* quote s constructs a regular expression that matches the same
     string as s, modular the number of spaces and newlines. *)

  let trim s =
    if Str.string_match (Str.regexp "^[ \t\n]+$") s 0
    then "[ \t\n]+"
    else if Str.string_match (Str.regexp "^[ \t\n]*\\([^ \t\n]+\\)[ \t\n]*$") s 0
    then "[ \t\n]*" ^ (Str.quote(Str.matched_group 1 s)) ^ "[ \t\n]*" 
    else invalid_arg ("unexpected delimiter: "^ s) in

  fun code ->
    String.concat ""
      (List.map (function
	| Str.Text t -> Str.quote t
	| Str.Delim d -> trim d
       ) (Str.full_split c_delimiter code))
      

let test_quote =
  "quote">::
    (fun _ -> 
      assert_bool "should match"
	(Str.string_match
	   (Str.regexp
	      (quote "hello   world.  \n  \t  are  \t \n\n you happy?"))
	   "hello world. are you\n\n  happy?    "
	   0);
      assert_bool "should not match"
	(not (Str.string_match
		(Str.regexp
		   (quote "hello   world.  \n  \t  are  \t \n\n you happy?"))
		"hello   world.  \n  \t  are  \t \n\n you ha ppy?"
		0)))

let test_quote2 =
  "quote">::
    (fun _ -> 
      assert_bool "should match"
	(Str.string_match
	   (Str.regexp
	      (quote "_f( x, y,gg(z))"))
	   "_f (x,y, gg ( z  )  )"
	   0);
      assert_bool "should not match"
	(not (Str.string_match
		(Str.regexp
	      (quote "_f( x, y,gg(z))"))
	   "_ f (x,y, g ( z  )  )"
		0)))

let test_quote3 =
  "quote3">::
    fun _->
      assert_bool "should contain"
	(0<=(Str.search_forward
	       (Str.regexp
		  (quote
"inline void __ikra_kernel_initialize0(int* ptr, int size) {
  for (int i = 0; i < size; ++(i))
    ptr[i] = __ikra_body_initialize0(i);
}"))
"VALUE initialize0wrapper(VALUE self, VALUE wrapped_array, VALUE _size) {
    __ikra_kernel_initialize0
     (get_parray(wrapped_array,int),
      NUM2INT(_size)
      ));
    return Qnil;
}
inline void __ikra_kernel_initialize0(int* ptr, int size) {
  for (int i = 0; i < size; ++(i))
    ptr[i] = __ikra_body_initialize0(i);
}" 0))

let is_contained : string -> string -> bool =
  (* is_contained s file determines if s appears in the file, ignoring
     whitespaces.  *)
  
  fun s file ->
    let in_file = Util.read_lines file in
    let pattern = Str.regexp (quote s) in
    try let _ = Str.search_forward pattern in_file 0 in
	true
    with Not_found _ -> false


let make_test_convert (source, wrapper, loop, body, initialize, preamble) =
  ("covert_" ^ source)>::
    fun _ ->
      Convert.target_file_names None "test_convert" (fun gen_env ->
	let cuda_file = (Convert_env.get_cuda_file_name gen_env) in
	(let t_source = 
	   (Ast.recognize
	      (Typeinf.recognize_type_decls
		 (Test_util.parse_ruby_string source))) in
	 (* Printf.printf "t_source=%s\n" *)
	 (*   (Ast.string_of ~show_type:true t_source); *)
	 let kernels = Convert.generate_ruby_file gen_env t_source in
	 try 
	   let _ = Convert.generate_cuda_file gen_env kernels in
	   assert_bool "generate_cuda_file should generate a file"
	     (Sys.file_exists cuda_file);
	   let assert_contains code =
	     assert_bool ("generated file should contain \n"^code
			  ^"\nbut:\n"^(Util.read_lines cuda_file))
	       (is_contained code cuda_file) in
	   assert_contains wrapper;
	   assert_contains loop;
	   assert_contains body;
	   assert_contains initialize;
	   assert_contains preamble
	 with Util.External_command_error(command, out, err) as e ->
	   Printf.printf "Error when run command: %s\n\nOutput:\n%s\n\n\
			   Errors:\n%s\n\nRuby:\n%s\n\nCUDA:\n%s\n"
	     command out err 
	     (Util.read_lines (Convert_env.get_ruby_file_name gen_env))
	     (Util.read_lines (Convert_env.get_cuda_file_name gen_env));
	   raise e))

let testdata_convert = [
  (* ********************************************************************** *)
  (* the body of a PArray call will be converted to a call on a kernel
     method. *)
  "_t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); x },[Array,1,Integer])",
  (* a wrapper function *)
  "\
VALUE __ikra_wrapper_initialize0(VALUE self,
		     VALUE _array,
		     VALUE _size0) {
    __ikra_kernel_initialize0(get_parray(_array, int),
			       NUM2INT(_size0));
    return Qnil;
 }",
  (* a loop function to be called from the wrapper, and to call the
     body function on each element *)
  "\
inline void __ikra_kernel_initialize0(int* _array, int _size0) {
  for (int _i0 = 0; (_i0) < (_size0); ++(_i0))
    (_array[((0)*(_size0))+(_i0)]) = (__ikra_body_initialize0(_i0));
}",
  (* a body function to compute something for each element *)
  "\
inline int __ikra_body_initialize0(int x) {
  return ({int _RETURN; x;});
}",
  "\
extern \"C\"
void Init_test_convert() {
  VALUE cKernel = rb_define_module(\"Ikra_Kernels__test_convert\");
  rb_define_module_function
    (cKernel,
     \"initialize0\",
     (VALUE(*)(...))(__ikra_wrapper_initialize0),
     2);
  return;
}",
"#include <ikra.h>"
;

  (* ********************************************************************** *)
  (* the kernel method call in the converted code should have
     lexical variables in the body as parameters. *)
  "_typedecl(Integer, :l); l=3; \
   _t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); x+l },[Array,1,Integer])",
  "\
VALUE __ikra_wrapper_initialize0(VALUE self,
		     VALUE _array,
		     VALUE _size0,\
		     VALUE fv0) {
    __ikra_kernel_initialize0(get_parray(_array, int),
			       NUM2INT(_size0),
			       NUM2INT(fv0));
    return Qnil;
 }",
  (* a loop function to be called from the wrapper, and to call the
     body function on each element *)
  "\
inline void __ikra_kernel_initialize0(int* _array, int _size0, int fv0) {
  for (int _i0 = 0; (_i0) < (_size0); ++(_i0))
    (_array[((0)*(_size0))+(_i0)]) = (__ikra_body_initialize0(_i0, fv0));
}",
  (* a body function to compute something for each element *)
  "\
inline int __ikra_body_initialize0(int x, int l) {
  return ({int _RETURN; (x)+(l);});
}",
  (* an initialization function that creates the interfacing module
     and that register the wrapper function. *)
  "\
extern \"C\"
void Init_test_convert() {
  VALUE cKernel = rb_define_module(\"Ikra_Kernels__test_convert\");
  rb_define_module_function
    (cKernel,
     \"initialize0\",
     (VALUE(*)(...))(__ikra_wrapper_initialize0),
     3);
  return;
}",
"#include <ikra.h>"
;

  (* ********************************************************************** *)
  (* floating point support *)
  "_typedecl(Float, :l); l=0.5; \
   _t(PArray.new(10){ |x| _typedecl(Integer, :x, Float, :_RETURN); x+l+0.5 },[Array,1,Float])",
  "\
VALUE __ikra_wrapper_initialize0(VALUE self,
		     VALUE _array,
		     VALUE _size0,\
		     VALUE fv0) {
    __ikra_kernel_initialize0(get_parray(_array, float),
			       NUM2INT(_size0),
			       IKRA_NUM2FLOAT(fv0));
    return Qnil;
 }",
  (* a loop function to be called from the wrapper, and to call the
     body function on each element *)
  "\
inline void __ikra_kernel_initialize0(float* _array, int _size0, float fv0) {
  for (int _i0 = 0; (_i0) < (_size0); ++(_i0))
    (_array[((0)*(_size0))+(_i0)]) = (__ikra_body_initialize0(_i0, fv0));
}",
  (* a body function to compute something for each element *)
  "\
inline float __ikra_body_initialize0(int x, float l) {
  return ({float _RETURN;((x)+(l))+(0.5);});
}",
  (* an initialization function that creates the interfacing module
     and that register the wrapper function. *)
  "\
extern \"C\"
void Init_test_convert() {
  VALUE cKernel = rb_define_module(\"Ikra_Kernels__test_convert\");
  rb_define_module_function
    (cKernel,
     \"initialize0\",
     (VALUE(*)(...))(__ikra_wrapper_initialize0),
     3);
  return;
}",
"#include <ikra.h>"
;

  (* ********************************************************************** *)
  (* test of using floating point numbers *)
  "_t(PArray.new(10){ |x| _typedecl(Integer, :x, Float, :_RETURN); \
			       x*0.1 },\
	   [Array, 1, Float]).\
	reduce(0.0){ |x,y| _typedecl(Float, :x, :y, :_RETURN); x+y }",
  (* a wrapper function *)
  "\
VALUE __ikra_wrapper_reduce0(VALUE self,
		     VALUE _array, VALUE _unit,
		     VALUE _size0) {
    return rb_float_new(__ikra_kernel_reduce0(\
	get_parray(_array, float), IKRA_NUM2FLOAT(_unit),
	NUM2INT(_size0)));
 }",

(*   "\ *)
(* VALUE __ikra_wrapper_initialize1(VALUE self,  *)
(* 		     VALUE _array, *)
(* 		     VALUE _size0) { *)
(*     __ikra_kernel_initialize1(get_parray(_array, float), *)
(* 			       NUM2INT(_size0)); *)
(*     return Qnil; *)
(*  }", *)

  (* a loop function *)
  "\
inline float __ikra_kernel_reduce0(float* _array, float _unit, int _size0) {\
  float accum = _unit;
  for (int _i0 = 0; (_i0) < (_size0); ++(_i0))
    (accum) =  (__ikra_body_reduce0(accum,_array[((0)*(_size0))+(_i0)]));
  return accum;
}",
(*   "\ *)
(* inline void __ikra_kernel_initialize1(float* _array, int _size0) { *)
(*   for (int _i0 = 0; _i0 < _size0; ++(_i0)) *)
(*     _array[((0)*(_size0))+(_i0)] = __ikra_body_initialize1(_i0); *)
(* }", *)
  (* a body function to compute something for each element *)
  "\
inline float __ikra_body_reduce0(float x, float y) {
  return ({float _RETURN; (x)+(y);});
}",
(*   "\ *)
(* inline float __ikra_body_initialize1(int x) { *)
(*   return (x)*(0.1); *)
(* }", *)
  (* don't care initialization and header *)
  "", "";





  (* (\* the lexical variables sholud be sorted in the alphabetical order. *\) *)
  (* "_typedecl(Integer, :l, :a, :b); b=4; l=3; a=9;\ *)
  (*  PArray.new(10){ |x| _typedecl(Integer, :x); a+x+l+b }", *)
  (* "_typedecl(Integer, :l, Integer, :a, Integer, :b); b=4; l=3; a=9;\ *)
  (*  PArray.new(10){ |_array,_size| Ikra_Kernel.initialize0(_array,_size,a,b,l) }"; *)

  (* (\* reduce calls on an array. *\) *)
  (* "_typedecl([Array,1,Integer], :a); a=0; a.reduce(0){ |x,y| x+y }", *)
  (* "_typedecl([Array,1,Integer], :a); a=0; \ *)
  (*  a.reduce(0){ |_array,_size,zero| Ikra_Kernel.reduce0(_array,_size,zero)}"; *)

  (* (\* inject is synonym of reduce *\) *)
  (* "_typedecl([Array,1,Integer], :a); a=0; a.inject(0){ |x,y| x+y }", *)
  (* "_typedecl([Array,1,Integer], :a); a=0; \ *)
  (*  a.reduce(0){ |_array,_size,zero| Ikra_Kernel.reduce0(_array,_size,zero)}"; *)

  (* (\* map on an array *\) *)
  (* "_typedecl([Array,1,Integer], :a); a=0; a.map{ |x| x*x }", *)
  (* "_typedecl([Array,1,Integer], :a); a=0; \ *)
  (*  a.map{ |_array,_size| Ikra_Kernel.map0(_array,_size)}"; *)

  (* (\* combination of new and map. *\) *)
  (* "_typedecl([Array,1,Integer], :a, Integer, :l, :m);\ *)
  (*  l = m = 1;\ *)
  (*  a=PArray.new(10){ |x| _typedecl(Integer, :x); x+l };\ *)
  (*  puts a.reduce(1){ |x,y|_typedecl(Integer, :x, :y); x*y*m }", *)
  (* "_typedecl([Array,1,Integer], :a, Integer, :l, Integer, :m);\ *)
  (*  l = m = 1;\ *)
  (*  a=PArray.new(10){ |_array,_size| Ikra_Kernel.initialize0(_array,_size,l) };\ *)
  (*  puts a.reduce(1){ |_array,_size,zero| Ikra_Kernel.reduce1(_array,_size,zero,m) }"; *)

  (* ********************************************************************** *)
  (* local variable inside a kernel *)
  "_t(PArray.new(10){ |x| _typedecl(Integer, :x, :y, :_RETURN); y=x; y },[Array,1,Integer])",
  (* a wrapper function *)
  "\
VALUE __ikra_wrapper_initialize0(VALUE self,
		     VALUE _array,
		     VALUE _size0) {
    __ikra_kernel_initialize0(get_parray(_array, int),
			       NUM2INT(_size0));
    return Qnil;
 }",
  (* a loop function to be called from the wrapper, and to call the
     body function on each element *)
  "\
inline void __ikra_kernel_initialize0(int* _array, int _size0) {
  for (int _i0 = 0; (_i0) < (_size0); ++(_i0))
    (_array[((0)*(_size0))+(_i0)]) = (__ikra_body_initialize0(_i0));
}",
  (* a body function to compute something for each element *)
  "\
inline int __ikra_body_initialize0(int x) {
  return ({int y; int _RETURN; (y)=(x);y;});
}",
  "\
extern \"C\"
void Init_test_convert() {
  VALUE cKernel = rb_define_module(\"Ikra_Kernels__test_convert\");
  rb_define_module_function
    (cKernel,
     \"initialize0\",
     (VALUE(*)(...))(__ikra_wrapper_initialize0),
     2);
  return;
}",
"#include <ikra.h>"
;

  (* ********************************************************************** *)
  (* a method definition called from a kernel *)
  "def m(x) _typedecl(Integer, :x, :_RETURN); x+0; end; _t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); m(x) }, [Array,1,Integer])",
  (* a wrapper function *)
  "\
VALUE __ikra_wrapper_initialize0(VALUE self,
		     VALUE _array,
		     VALUE _size0) {
    __ikra_kernel_initialize0(get_parray(_array, int),
			       NUM2INT(_size0));
    return Qnil;
 }",
  (* a loop function to be called from the wrapper, and to call the
     body function on each element *)
  "\
inline void __ikra_kernel_initialize0(int* _array, int _size0) {
  for (int _i0 = 0; (_i0) < (_size0); ++(_i0))
    (_array[((0)*(_size0))+(_i0)]) = (__ikra_body_initialize0(_i0));
}",
  (* a body function to compute something for each element *)
  "\
inline int __ikra_body_initialize0(int x) {
  return ({int _RETURN;m(x);});
}",
  "\
extern \"C\"
void Init_test_convert() {
  VALUE cKernel = rb_define_module(\"Ikra_Kernels__test_convert\");
  rb_define_module_function
    (cKernel,
     \"initialize0\",
     (VALUE(*)(...))(__ikra_wrapper_initialize0),
     2);
  return;
}",
"int m(int x) {
  return ({int _RETURN;(x)+(0);});
}"
;

  (* zip support *)
  "_typedecl([Array,1,[Tuple,Integer,Integer]],:a);\
   a=0;
   _t(_t(a,[Array,1,[Tuple,Integer,Integer]])\
      .map{ |x,y| _typedecl(Integer, :x, :y, :_RETURN); x+y },\
      [Array,1,Integer]);",
  "VALUE __ikra_wrapper_map0(VALUE self,VALUE _array_t0,VALUE _array_t1,\
                             VALUE _array2,VALUE _size0){
     __ikra_kernel_map0(get_parray(_array_t0,int),\
                        get_parray(_array_t1,int),\
                        get_parray(_array2,int),\
                        NUM2INT(_size0));
     return Qnil;
   }",
  "inline void __ikra_kernel_map0(int* _array_t0,int* _array_t1,
                                  int* _array2,int _size0){
     for(int _i0=0;(_i0)<(_size0);++(_i0))
       (_array2[((0)*(_size0))+(_i0)])
        =(__ikra_body_map0(_array_t0[((0)*(_size0))+(_i0)],
                           _array_t1[((0)*(_size0))+(_i0)]));
   }",
  "inline int __ikra_body_map0(int x,int y){
     return ({int _RETURN; (x)+(y);});
   }",
  "extern \"C\"
   void Init_test_convert(){
       VALUE cKernel=rb_define_module(\"Ikra_Kernels__test_convert\");
       rb_define_module_function(cKernel,\"map0\",
                                 (VALUE(*)(...))(__ikra_wrapper_map0),4);
       return;
   }",
  "// This file is generated for test_convert.
   #include <ikra.h>";

(* neighbor support *)
  "_typedecl([Array,1,[Neighbor,3,1,Integer]],:a);\
   a=0;
   _t(_t(a,[Array,1,[Neighbor,3,1,Integer]])\
      .map{ |m| _typedecl([Neighbor,3,1,Integer], :m, 
                          Integer, :_RETURN); 
        _t(m,[Neighbor,3,1,Integer])[-1]*
        _t(m,[Neighbor,3,1,Integer])[0]*
        _t(m,[Neighbor,3,1,Integer])[1] },\
      [Array,1,Integer]);",
  "VALUE __ikra_wrapper_map0(VALUE self,VALUE _array_array,VALUE _array_pad,\
                             VALUE _array2,VALUE _size0){
     __ikra_kernel_map0(get_parray(_array_array,int),NUM2INT(_array_pad),\
                        get_parray(_array2,int),NUM2INT(_size0));
     return Qnil;
   }",
  "inline void __ikra_kernel_map0(int* _array_array,int _array_pad,\
                                  int* _array2,int _size0){
     for(int _i0=0;(_i0)<(_size0);++(_i0))
       (_array2[((0)*(_size0))+(_i0)]) =
         (__ikra_body_map0(
           safe_aref1(_array_array,(_i0)+(-1),_size0,_array_pad),\
           safe_aref1(_array_array,(_i0)+(0),_size0,_array_pad),\
           safe_aref1(_array_array,(_i0)+(1),_size0,_array_pad)));
  }",
  "inline int __ikra_body_map0(int m$_1,int m$0,int m$1){
        return ({int _RETURN; ((m$_1)*(m$0))*(m$1);});
   }",
  "extern \"C\"
   void Init_test_convert(){
     VALUE cKernel=rb_define_module(\"Ikra_Kernels__test_convert\");
    rb_define_module_function(cKernel,\"map0\",\
                              (VALUE(*)(...))(__ikra_wrapper_map0),4);
    return;
  }",
  "#include <ikra.h>";

  (* constant support *)
  "_typedecl(Integer, :A_CONST);\
   _t(A_CONST=1, Integer); _t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); A_CONST },[Array,1,Integer])",
  "int A_CONST = 1;",
  "",
  "",
  "",
  ""
  ;

  (* constant support 2 *)
  "_typedecl(Integer, :A_CONST, :ANOTHER_CONST);\\
   def m() _typedecl(Integer, :_RETURN); 123; end
   _t(ANOTHER_CONST=m(),Integer);
   _t(A_CONST=ANOTHER_CONST+1,Integer); _t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); A_CONST },[Array,1,Integer])",
  "const int ANOTHER_CONST = m();
   const int A_CONST = (ANOTHER_CONST)+(1);",
  "int m() { return ({int _RETURN;123;}); }",
  "",
  "",
  ""
  ;

  (* return statement *)
  "def m(n) 
     _typedecl(Integer, :_RETURN, :n)
     if n==0 then return 0; end
     1
   end
   _t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); m(x) },
      [Array,1,Integer])",
  "int m(int n){
     return ({int _RETURN;if((n)==(0))return 0; else {}1;});
   }",
  "", "", "", "";
  
(* && and || in gather *)
  "def m() _typedecl(Integer, :_RETURN);\
     if !(true && false) || true then 0 else 1 end; end;\
   _t(PArray.new(10){ |x| _typedecl(Integer, :x, :_RETURN); m() },
      [Array,1,Integer])",
  "int m(){
     return ({int _RETURN;((!((TRUE)&&(FALSE)))||(TRUE))?(0):(1);});
   }",
  "", "", "", "";

  (* tuple returning kernel *)
  "_t(PArray.new(10){ |x| _typedecl(Integer, :x, 
                                    [Tuple,Integer,Integer], :_RETURN); 
      _t([x,x],[Tuple,Integer,Integer]) },
      [Array,1,[Tuple,Integer,Integer]])",
  "struct ikra_tuple_2II {int f0;int f1;};\
inline struct ikra_tuple_2II ikra_tuple_2II(int f0, int f1) {
  struct ikra_tuple_2II s;
  ((s).f0) = (f0); 
  ((s).f1) = (f1);
  return s;
}",
  "inline struct ikra_tuple_2II __ikra_body_initialize0(int x){
    return ({struct ikra_tuple_2II _RETURN;ikra_tuple_2II(x,x);});
   }",
  "inline void __ikra_kernel_initialize0(int* _array_t0,int* _array_t1,int _size0){
     for(int _i0=0;(_i0)<(_size0);++(_i0)) {
        struct ikra_tuple_2II r = __ikra_body_initialize0(_i0);
        (_array_t0[((0)*(_size0))+(_i0)])=((r).f0);
        (_array_t1[((0)*(_size0))+(_i0)])=((r).f1);
     }
   }",
  "VALUE __ikra_wrapper_initialize0(VALUE self,VALUE _array_t0,
                                    VALUE _array_t1,VALUE _size0){
     __ikra_kernel_initialize0(get_parray(_array_t0,int),
                               get_parray(_array_t1,int),
                               NUM2INT(_size0));
     return Qnil;
   }",
  "extern \"C\"
   void Init_test_convert(){
     VALUE cKernel=rb_define_module(\"Ikra_Kernels__test_convert\");
     rb_define_module_function(cKernel,\"initialize0\",
                               (VALUE(*)(...))(__ikra_wrapper_initialize0),3);
     return;
   }";
  
    (* tuple returning map *)
  "_t(_t(a,[Array,1,Integer]).map{ |x| 
             _typedecl(Integer, :x, [Tuple,Integer,Integer], :_RETURN);
             _t([x,x],[Tuple,Integer,Integer]) 
            },
      [Array,1,[Tuple,Integer,Integer]])",
  "struct ikra_tuple_2II {int f0; int f1;};
   inline struct ikra_tuple_2II ikra_tuple_2II(int f0,int f1){
        struct ikra_tuple_2II s;
        ((s).f0)=(f0); ((s).f1)=(f1); return s;
   }",
  "inline struct ikra_tuple_2II __ikra_body_map0(int x){
        return ({struct ikra_tuple_2II _RETURN;ikra_tuple_2II(x,x);});
   }",
  "inline void __ikra_kernel_map0(int* _array,int* _array2_t0,
                                  int* _array2_t1,int _size0){
     for(int _i0=0;(_i0)<(_size0);++(_i0)){
       struct ikra_tuple_2II r=__ikra_body_map0(_array[((0)*(_size0))+(_i0)]);
       (_array2_t0[((0)*(_size0))+(_i0)])=((r).f0);
       (_array2_t1[((0)*(_size0))+(_i0)])=((r).f1);
     }
   }",
  "VALUE __ikra_wrapper_map0(VALUE self,VALUE _array,VALUE _array2_t0,
                             VALUE _array2_t1,VALUE _size0){
      __ikra_kernel_map0(get_parray(_array,int),get_parray(_array2_t0,int),
                         get_parray(_array2_t1,int),NUM2INT(_size0));
      return Qnil;
   }",
  "";


  (* auxiliary functions and constants *)
  "_typedecl(Integer, :CONST);
   _t(CONST=1,Integer);
   def f(x,y) _typedecl(Integer, :x, Integer, :y, Integer, :_RETURN);
      x+y 
   end;
   _t(_t(a,[Array,1,Integer]).map{ |x| 
             _typedecl(Integer, :x, Integer, :_RETURN);
             f(x,CONST)
            });",
  "int CONST = 1;",
  "int f(int x,int y){ return ({int _RETURN;(x)+(y);}); }",
  "",
  "",
  ""
  ;

]

let test_convert_exps = List.map make_test_convert testdata_convert

(* let make_test_convert1 (source, wrapper, loop, body, initialize, preamble) = *)
(*   ("covert_" ^ source)>:: *)
(*     fun _ -> *)
(*       Convert.target_file_names None "test_convert1" (fun gen_env -> *)
(* 	let cuda_file = (Convert_env.get_cuda_file_name gen_env) in *)
(* 	(let t_source =  *)
(* 	   (Ast.recognize *)
(* 	      (Typeinf.recognize_type_decls *)
(* 		 (Test_util.parse_ruby_string source))) in *)
(* 	 (\* Printf.printf "t_source=%s\n" *\) *)
(* 	 (\*   (Ast.string_of ~show_type:true t_source); *\) *)
(* 	 let kernels = Convert.generate_ruby_file gen_env t_source in *)
(*          Printf.printf "# kernels = %d\n" (List.length kernels); *)
(*          (assert_bool *)
(*             (Printf.sprintf "there should be at least one kernel in %s" *)
(*                source) *)
(*             ((List.length kernels) != 0)))) *)

(* let test_convert_exps1 = List.map make_test_convert1 testdata_convert *)


let make_test_top_decls (source, expected_method_names) =
  ("test_top_decls:"^source)>::(fun _ ->
    Convert.target_file_names None "test_convert" (fun gen_env ->
      (* let cuda_file = (Convert_env.get_cuda_file_name gen_env) in *)
      (let t_source = 
	 (Ast.recognize
	    (Typeinf.recognize_type_decls
	       (Test_util.parse_ruby_string source))) in
       let kernels = Convert.generate_ruby_file gen_env t_source in
       let decls = Convert_kernel.top_decls kernels in
       let actual_method_names =
         List.concat (List.map
                        (function (Ast.Method_def(n,_,_),_) -> [n] | _ -> [])
                        decls) in
       let set_equal set1 set2 =
         (List.for_all (fun n -> List.mem n set2) set1) &&
         (List.for_all (fun n -> List.mem n set1) set2)  in
       assert_equal ~cmp:set_equal ~printer:(fun decls ->
         (String.concat ";" decls))
         expected_method_names actual_method_names)))
    
let test_top_decls = List.map make_test_top_decls [
  "_t(Array.new(1){|x|_typedecl(Integer,:x);m()},[Array,1,Integer]);def m() _typedecl(Integer,:_RETURN); 0 end", ["m"];
  "_t(Array.new(1){|x|_typedecl(Integer,:x);m()},[Array,1,Integer]);def m() _typedecl(Integer,:_RETURN); 0 end; def  n() _typedecl(Integer,:_RETURN);0 end", ["m";"n"];
  "_typedecl(Integer, :a); _t(Array.new(1){|x|_typedecl(Integer,:x);x},[Array,1,Integer]);a=0; def m() _typedecl(Integer,:_RETURN); 0 end; def n() _typedecl(Integer,:_RETURN); 0 end", ["m";"n"];
]


let suite =
  "suite">:::[
    test_quote;
    test_quote2;
    test_quote3;
  ]
    @ test_convert_exps
    @ test_top_decls
    (* @test_convert_exps1 *)

let _ = Config.initialize None

let () =
  run_test_tt_main suite
