open OUnit2

(* This test program checks whether the frontend compiler can create
   appropriate build and converted ruby scripts.

   We require that 

   1. ikrac can build everything and run the compiled code,

   2. ikrac generates a build script, which compiles and automatically
      executes, and

   3. ikrac generates a build script, which compiles converted
      programs into executables, which can be executed.

   We also require that the execution of the converted Ruby program
   can be done at a different directory from the directory of the
   converted program.

   Hence it will test compiling/building/running in different ways.
*)

let with_temp_dir : (string -> 'a) -> 'a =
  (* create a temporary direcotry, chdir to there, run a function, and
     erase the directory *)
  fun body -> 
    let tmp_dir_name = Util.make_temp_dir "ikra_test" in
    Util.with_finally (fun _ -> 
      (Util.with_cd tmp_dir_name 
         (fun _ -> body tmp_dir_name)))
      (fun _ ->
        let _ = Util.run_command ["/bin/rm"; "-r"; tmp_dir_name] in
        ())

let build_output result =               
  (* expected output from build.sh where the temporary names are
     matched with REGEXP patterns*)
  String.concat "" ("creating Makefile" :: result)

let cmp_rexep expected_regexp actual =
  let re = Str.regexp expected_regexp in
  Str.string_match re actual 0

let target_dir_name = "gen"
let test_program = "puts Array.new(10){|x| x}.reduce(0){|x,y|x+y}"
let expected_output = "45"

type body_env = {
  rubyfile : string;
  target_dir_name : string;
  compiler_output : string;
}

let converted_ruby_file ruby_file = 
  Filename.basename
    (String.concat "." [Filename.chop_suffix ruby_file ".rb"; "ikra.rb"])
    
let compile_and_exec name generate_only body = name >:: (fun _ ->
  with_temp_dir (fun tmpdir ->
    Unix.mkdir target_dir_name 0o755;
    Util.with_temp_file (Filename.concat tmpdir "test") ".rb"
      test_program
      (fun rubyfile ->
        let compiler_output = 
          Frontend.compile_file ~cleanup:false ~generate_only
            ~target_dir:(Some target_dir_name) ~redirect:true 
            rubyfile [] in
        body {rubyfile; target_dir_name; compiler_output})))

(* the following four test cases test compilation, buliding, and
   execution in different situations *)

let compile_auto_build =                
  (* ikrac --target-dir gen test.rb  *)
  compile_and_exec "compile_auto_build" false 
    (fun {compiler_output} ->
      assert_equal ~printer:(fun s->s)  ~cmp:cmp_rexep
        expected_output compiler_output)

let build_auto_run = 
  (* ikrac --target-dir gen test.rb
     cd gen; sh build.sh                *)
  compile_and_exec "build_then_run_in_diff_dir" true
    (fun {rubyfile; target_dir_name} ->
    Util.with_cd target_dir_name (fun _ ->
      assert_equal ~printer:(fun s->s)  ~cmp:cmp_rexep
        (build_output ["\n"; expected_output])
        (Util.run_command ["sh"; "build.sh";])))

let build_then_run_in_diff_dir =  
  (* ikrac --target-dir gen test.rb
     cd gen; sh build.sh --no-exec
     cd ..; ruby gen/test.ikra.rb
  *)
  compile_and_exec "build_then_run_in_diff_dir" true
    (fun {rubyfile; target_dir_name} ->
      (Util.with_cd target_dir_name (fun _ ->
        assert_equal ~printer:(fun s->s)  ~cmp:cmp_rexep
          (build_output [])
          (Util.run_command ["sh"; "build.sh"; "--no-exec"]));
       assert_equal ~printer:(fun s->s) 
         expected_output
         (Util.run_ruby [
           (Filename.concat
              target_dir_name (converted_ruby_file rubyfile))])))

let build_then_run_in_same_dir =  
  (* ikrac --target-dir gen test.rb
     cd gen; sh build.sh --no-exec
     ruby test.ikra.rb
  *)
  compile_and_exec "build_then_run_in_diff_dir" true
    (fun {rubyfile; target_dir_name} ->
      (Util.with_cd target_dir_name (fun _ ->
        assert_equal ~printer:(fun s->s)  ~cmp:cmp_rexep
          (build_output [])
          (Util.run_command ["sh"; "build.sh"; "--no-exec"]);
        assert_equal ~printer:(fun s->s) 
          expected_output
          (Util.run_ruby [(converted_ruby_file rubyfile)]))))

let suite =
  "suite">:::
    [compile_auto_build;
     build_auto_run;
     build_then_run_in_diff_dir;
     build_then_run_in_same_dir]

let _ = Config.initialize None

let () =
  run_test_tt_main suite

