open OUnit2
open Parser
open Ast
open Ikra_type
open Ast.Untyped_constructor

let _ = Config.initialize None          (* need for using functions in config *)

let string_of_ast = (fun exp -> Ast.string_of (exp,Ikra_type.Tobject))

let test_parser _ =
  let exp,_ = Test_util.parse_ruby_string "123" in
  assert_equal ~printer:string_of_ast
    (Ast.Literal (Integer_literal 123)) exp

let parsing_test_data = [
  "-1",literal (Integer_literal (-1));
  "123", literal (Integer_literal 123);
  "1+2", call(literal (Integer_literal 1), "+",
	      [literal (Integer_literal 2)], None);
  "puts 1+2", call(Ast.implicit_self, "puts", [call(literal (Integer_literal 1),
				       "+",
				       [literal (Integer_literal 2)],
				       None)],
		   None);
  "Integer", constant "Integer";
  "A", constant  "A";
  "x=1", local_assign("x", (literal (Integer_literal 1)));
  "x", call(Ast.implicit_self, "x", [], None);
  "x=1;x", block([local_assign("x", (literal (Integer_literal 1)));
		  local_var "x"]);
  "m(){ |x, y| 0 }", call(Ast.implicit_self, "m", [],
			  Some(proc(array_pattern[var_pattern("x");
						   var_pattern("y")],
				     literal (Integer_literal 0))));
  "m(){ |x| 0 }", call(Ast.implicit_self, "m", [],
		       Some(proc(var_pattern("x"),
				 literal (Integer_literal 0))));
  "m(){  0 }", call(Ast.implicit_self, "m", [],
		    Some(proc(empty_pattern, literal (Integer_literal 0))));
  "def m(x); x; end", method_def("m", array_pattern[var_pattern("x")],
				 (block[local_var "x"]));
  ":x", literal (Symbol_literal "x");
  "m(){ |x| x; 0 }",
  call(Ast.implicit_self, "m", [],
       Some(proc(var_pattern ("x"), block[local_var "x";
				          literal (Integer_literal 0)])));
  "anArray=anArray[123]",
  local_assign("anArray",
	       call(local_var "anArray", "[]",
		    [literal (Integer_literal 123)], None));
  "self", local_var "self";
  (* "_typedecl([Array,1,Integer], :a, Integer, :l, :m);\ *)
  (*  l = m = 1;\ *)
  (*  a=PArray.new(10){ |x| _typedecl(Integer, :x); x+l };\ *)
  (*  puts a.reduce(1){ |x,y|_typedecl(Integer, :x, :y); x*y*m }", *)
  (* Constant("dummy") *)

  "Math::PI", constant "Math::PI";

  "$0", constant "$0";

  "m(0,&:+)", call(Ast.implicit_self, "m", [literal (Integer_literal 0)],
                   Some(literal (Symbol_literal "+")));

]
  
let more_test_parser =
  List.map (fun (source, (expected)) ->
    "parse "^source >::
      (fun _ ->
	let actual = Test_util.parse_ruby_string source in
	assert_equal ~printer:Ast.string_of expected actual))
    parsing_test_data

(* As you can see in the discussion at
   http://stackoverflow.com/questions/9911781/ruby-global-scope, "p
   123" and "self.p 123" are different in Ruby.  When the method is
   private, the former fails even if it is made in the same scope of
   that class.  Therefore, in a strict sense, it is dangerous to
   complement "self" automatically.  We should perhaps to complement
   with another pesudo variable (like "**implicit_self**") and erase
   it when unparsing. 
*)
let parse_unparse_data = [
  "p 123";
  "x = 1; p x";
  "a = Array.new(10){|x| x }; p a.reduce(0){|x,y| x+y }";
  "p 'hello'";
  "p 1.23";
  "def m(); 0; end";
  "def m(); yield [1,[2,3],4]; end ; p m(){ |x,(y,z),w| x+y+z+w }";
  "def m(x); x; end; p m(123)";
  "i=10;while i > 0 do p i-=1; end";
  "if 3 <= 4 then p 123 else p 999 end";
  "if 3 <= 4 then p 123 end";
  "for i in 1..3 do p i; end";
  "k=3;for i in 1..k do p i; end";
  "p :symbol; p :symbol.class";
  "a = Array.new(1){ 123}; p a[0]";
  "1.0e-06";
  "C=1";
  "for i in 1..10 do; puts i; end";
  "def m() return 1; end";
  "def m() while true do return; end; end";
]

let parse_to_mlexp : (string * string) list =
  (* to represent source and parsed programs where parsed programs are
     given as a string of OCaml expressions *)
  [
    "0", "Literal(Integer_literal 0)";
    
    "unknownMethod(){ |x| anotherUknown(){ 1 } }", 
    "Call(Local_var(\"**self**\"),\"unknownMethod\",[],\
        Some(Proc(Var_pattern(\"x\"),\
                Call(Local_var(\"**self**\"),\"anotherUknown\",[],\
                    Some(Proc(Array_pattern[],Literal(Integer_literal 1)))\
                   )\
            ))\
       )";
  ]

let tests_mlexp =
  List.map (fun (src, expected) ->
    "parse_to_mlexp:"^src >::
      (fun _ ->
        let exp = Test_util.parse_ruby_string src in
        let actual = Ast.mlexp_of Ikra_type.string_of exp in
        assert_equal ~printer:(fun x->x)
          expected actual))
    parse_to_mlexp

let run_ruby_script : string -> string =

  Test_util.with_string 
    (fun filename -> Util.run_ruby [filename])

let test_by_parse_unparse_eval =
  List.map (fun source ->
    "parse_unparse "^source >::
      (fun _ ->
	let exp = Test_util.parse_ruby_string source in
	let unparsed = Ast.string_of exp in
	assert_equal ~printer:(fun x->x)
	  (run_ruby_script source)
	  (try run_ruby_script unparsed
	   with Invalid_argument msg ->
	     (raise (Invalid_argument (msg ^ " while executing unparsed: "
				       ^ unparsed))))
      ))
    parse_unparse_data

let ruby_parser_source = "puts 0+(-3)"
let ruby_parser_expected = "s(:call, nil, :puts, s(:arglist, s(:call, s(:lit, 0), :+, s(:arglist, s(:lit, -3)))))"
let ruby_parser_tokens = ["s"; "("; ":"; "call"; ","; "nil"; ","; ":"; "puts"; ",";  "s"; "("; ":"; "arglist"; ","; "s"; "("; ":"; "call"; ","; "s"; "("; ":"; "lit"; ","; "0"; ")"; ","; ":"; "+"; ",";  "s"; "("; ":"; "arglist"; ","; "s"; "("; ":"; "lit"; ","; "-3"; ")"; ")"; ")"; ")"; ")"; ]
let ruby_parser_sexp = 
  Sexp_list 
    [Sexp_symbol "call";
     Sexp_nil;
     Sexp_symbol "puts";
     Sexp_list
       [Sexp_symbol "arglist";
	Sexp_list [Sexp_symbol "call";
		   Sexp_list [Sexp_symbol "lit"; Sexp_int 0];
		   Sexp_symbol "+";
		   Sexp_list [Sexp_symbol "arglist";
			      Sexp_list [Sexp_symbol "lit"; Sexp_int (-3)]]]]]

let test_ruby_parser _ =
  assert_equal ~printer:(fun s->s) ruby_parser_expected
    (Test_util.with_string Parser.ruby_parse_file ruby_parser_source)

let test_sexp_parser _ =
  assert_equal
    ruby_parser_sexp ~printer:Parser.string_of_sexp
    (Parser.parse_sexp_string ruby_parser_expected)

let test_sexp_tokenize_sexp _ =
  assert_equal ruby_parser_tokens
    (Parser.tokenize_sexp ruby_parser_expected)

let test_sexp_tokenize_sexp2 _ =
  assert_equal ["s";"(";"nil";")"]
    (Parser.tokenize_sexp "s(nil)")

let test_sexp_tokenize_sexp3 _ =
  assert_equal ["s";"(";"\"abc\"";")"]
    (Parser.tokenize_sexp "s(\"abc\")")

let test_sexp_tokenize_sexp4 _ =
  assert_equal ~printer:(fun x->"["^(String.concat ";" x)^"]")
    ["dot2"]
    (Parser.tokenize_sexp "dot2")

let test_sexp_tokenize_sexp5 _ =
  assert_equal ~printer:(fun x->"["^(String.concat ";" x)^"]")
    ["s";"(";":";"lit";",";"-3";")"]
    (Parser.tokenize_sexp "s(:lit,-3)")

let test_sexp_tokenize_sexp6 _ =
  assert_equal ~printer:(fun x->"["^(String.concat ";" x)^"]")
    ["s";"(";":";"gvar";",";":"; "$0";")"]
    (Parser.tokenize_sexp "s(:gvar, :$0)")

let test_sexp_tokenize_sexp7 _ =
  assert_equal ~printer:(fun x->"["^(String.concat ";" x)^"]")
    ["s";"(";":";"lit";","; "1.0e-06";")"]
    (Parser.tokenize_sexp "s(:lit, 1.0e-06)")

let test_sexp_tokenize_sexp8 _ =
  assert_equal ~printer:(fun x->"["^(String.concat ";" x)^"]")
    ["s";"(";":";"call";","; "s";"(";":";"lit";",";"1";")";","; ":";"**";","; "s";"(";":";"arglist";","; "s";"(";":";"lit";","; "2";")";")";")"]
    (Parser.tokenize_sexp "s(:call, s(:lit, 1), :**, s(:arglist, s(:lit, 2)))")

let test_parse_1 _ =
  assert_equal (Sexp_int (-1),[]) (Parser.parse_sexp ["-1"])

let test_unparse_proc_param _ =
  let (exp) = (proc(array_pattern[var_pattern ("x");
				   var_pattern ("y")],
		     literal (Integer_literal 0))) in
  assert_equal ~printer:(fun x->x) " | x,y | 0"
    (Ast.string_of exp)

let suite =
  "suite">:::[
    "test_parser"                >:: test_parser; 
    "test_ruby_parser"           >:: test_ruby_parser;
    "test_sexp_parser"           >:: test_sexp_parser;
    "test_sexp_tokenize_sexp"    >:: test_sexp_tokenize_sexp;
    "test_sexp_tokenize_sexp2"   >:: test_sexp_tokenize_sexp2;
    "test_sexp_tokenize_sexp3"   >:: test_sexp_tokenize_sexp3;
    "test_sexp_tokenize_sexp4"   >:: test_sexp_tokenize_sexp4;
    "test_sexp_tokenize_sexp5"   >:: test_sexp_tokenize_sexp5;
    "test_sexp_tokenize_sexp6"   >:: test_sexp_tokenize_sexp6;
    "test_sexp_tokenize_sexp7"   >:: test_sexp_tokenize_sexp7;
    "test_sexp_tokenize_sexp8"   >:: test_sexp_tokenize_sexp8;
    "test_parse_1"               >:: test_parse_1;
    "test_unparse_proc_param"	 >:: test_unparse_proc_param;
  ] @ more_test_parser
  @ test_by_parse_unparse_eval
  @ tests_mlexp

let () =
  run_test_tt_main suite
;;
