(* test to do type inference and then resolve generated constraints  *)

open OUnit2
open Type_constraints			(* imports Consraints and Constr *)

let trace = false (* a switch to see a trace of constraint resolution *)

let parse_and_add_tvar : string -> Ast.tast = 
  fun source ->
    let source_exp = Test_util.parse_ruby_string source in
    let source_with_tdecl =
      Ast.recognize(Let_intro.let_intro source_exp) in
    Typeinf.add_type_vars source_with_tdecl 

let do_typeinf : string -> Constraints.t =
  fun source ->
    let source_with_tvar = (parse_and_add_tvar source) in
    (if trace then
	print_string ((Ast.string_of ~show_type:true source_with_tvar)^"\n")
     else ();
     Typeinf.generate_constraints source_with_tvar)

let cresolve : Constraints.t ->
  (Constraints.t * Cresolve.subst_t) =
  Cresolve.resolve

let set_print = fun s -> "{" ^ 
  (String.concat "," (List.map Constr.string_of
                        (Constraints.elements s))) ^ 
  "}"

let printer = fun (c,s) -> 
  Printf.sprintf "c=%s, s=[%s]" (set_print c)
    (String.concat ","
       (List.map (fun (t1,t2) ->
         (Ikra_type.string_of t1) ^ "->" ^ (Ikra_type.string_of t2))
          s))

let cmp = fun (expected_craints,s1) (actual_craints,s2) ->
  Constraints.equal
    (Type_constraints.remove_method_defs expected_craints)
    (Type_constraints.remove_method_defs actual_craints)

module ExpectedTypeConstraintConstructors = struct
  open Ikra_type
  let (<) t0 t1 = Constr.Less_than(t0,t1)
  let (=) t0 t1 = Constr.Equal(t0,t1)
  let t n = Tvar n
  let cset =
    List.fold_left (fun set con -> Constraints.add con set)
      Constraints.empty
  let mtype receiver name args return =
    Constr.Has_method(receiver,name,Tproc(args,return,None))
end

let parse_inference_resolve_check source =
  ("parse_inference_resolve:"^source)>::
    (fun _ ->
      assert_equal ~printer ~cmp 
 	(Constraints.empty,[])
	(cresolve (do_typeinf source)))

let parse_inference_unresolve_check source =
  ("parse_inference_unresolve:"^source)>::
    (fun _ ->
      assert_equal ~printer ~cmp:(fun expected actual ->
        not (cmp expected actual))
 	(Constraints.empty,[])
	(cresolve (do_typeinf source)))

let success_suite =
  (List.map parse_inference_resolve_check [
    (* These are the test cases of type inferencing.  An assertion is
       given as an expression like exp0.type_assert(exp1), which tests
       inferenced type of exp0 is equal to the one of exp1. *)

    (* Type of x can be either Int, Float, or even Object; but we want
       to fix it as Int. *)
    "x=1; x=x+1; x.type_assert(0)";

    (* The first assginment might make the type of x as Int, but
       should be Float since the RHS of the second statement yields a
       Float value, which is assgined to y.  For this inference, we
       need to have a rule "when a float value is multiplied to a
       value of a supertype of Int, the result will be float." *)
    "x=1; x=x+1.0; x.type_assert(1.0)";

    (* just to test two variables *)
    "x=1; y=x; (x+y).type_assert(0)";

    (* two variables interacting but have different types *)
    "x=1; y=x+1.0; x.type_assert(0); y.type_assert(1.0)";

    (* array support: new and reference with 1d-array *)
    "x=Array.new(10){ |i| i.type_assert(0); 1.23 }; x[0].type_assert(1.0)";

    (* the next case uses a block that omits its parameter; this is
       well-defined in Ruby, but not so important at this point. *)
    (* "x=Array.new(10){ 0 }; x[0].type_assert(0)"; *)

    (* array support: map *)
    "a = Array.new(100){ |i| i*i*1.0 }; b = a.map{ |v| v/2 };
    	b.type_assert(Array.new(1){ |i| 1.0 })";

    (* array support: reduce *)
    "Array.new(100){ |i| i }.reduce(0){ |x,y| x+y+0.0 }.type_assert(0.0)";

    (* method support *)
    "def m(x) x+1; end; puts m(123).type_assert(0)";

    (* (mutually) recursive calls *)
    "\
def m(x) if x==0 then 0 else n(x-1) end; end
def n(x) if x==0 then 1 else m(x-1) end; end
puts Array.new(10){|x| m(x)}.reduce(0){ |x,y| x+y }";

    "\
def m(x) if x==0 then 0 else n(x-1) end; end
def n(x) if x==0 then 1 else m(x-1) end; end
puts m(0);";

    "\
def m(x) if x==0 then 0 else m(x-1) end; end
puts m(0);";

    (* This does not work as we can't write a literal of an array of
       neighbors. *)
    (* multi dimensional array *)
    (* "Array.new_(10,10){ |x,y| x*10+y }.\ *)
    (*  type_assert( _t(0,[Array,2,Integer]))"; *)

    (* "Array.new_(10,10){ |x,y| x*10+y }.neighbor9(0).\ *)
    (* type_assert(_t(0,[Array,2,[Neighbor,9,2,Integer]]))"; *)

    (* Math functions and constants *)
    "Math::PI.type_assert(0.0);\
     x=Math::cos(x).type_assert(0.0);\
     x=Math.sin(x).type_assert(0.0)";

    (* if-then-without-else *)
    "(if 0==0 then 123 end).type_assert(puts 0)";

    (* boolean constants *)
    "true.type_assert(0==0);
     false.type_assert(0==0);";

    (* range expressions *)
    "a=0;(0..1).type_assert(a..a)";

    (* for expression *)
    "for i in 1..10 do; i.type_assert(0); end.type_assert(1..10)";
    "for i in 1..10 do; puts i; end; i.type_assert(0)";

    (* block parameter *)
    "Array.new_(3){ |x| x}.reduce(&:+)";
    (* "Array.new_(3,&:to_f).map{ |x| x/2 }.reduce(&:+)"; *)

    (* return statement *)
    "def m() if true then return 0.5; end; 1; end;
     Array.new_(3){ |x| m().type_assert(0.0) }.reduce(0){|x,y| x+y}.type_assert(0.0)";

    "def m(n)
      if n%2==0 then return 0.5; end
      1
     end
    puts Array.new_(5){ |n| m(n).type_assert(0.0) }.reduce(0){ |x,y| x+y }";

    (* a call on an unknown method should be allowed, and can take any
       parameters including block parameters *)
    "unknownMethod(){ 123 }";
    "UnknownClass.unknownMethod(){ |x| anotherUknown(){ 1 } }";
    "unknownMethod(){ |x| anotherUknown(){ 1 } }";
    "unknownMethod(){ 1 }";

  (* uses arithmetic operators on unknown classes *)
    "start=Time.now; Array.new(10){ |x| x }; puts Time.new - start";
    "puts 'hello' + 123.to_s + 'world'"; (* FIXME *)

  (* boolean conjunctions and disjunctions *)
    "(false && true).type_assert(true)";
    "(false || true).type_assert(true)";
    "(!false).type_assert(true)";

    (* boolean used in a conditional branch *)
    "n=1;(if true && n<2 then n else 0 end).type_assert(0)";

    (* array of tuples *)
    "Array.new(1){|i| i }.map{ |i| [i+1,i+2] }.map{ |x,y| x*y }"
  ])

let failure_suite = 
  (* to represent programs that cannot succeed in type inference;
     i.e., those that cause type errors *)
  (List.map parse_inference_unresolve_check [

    "UnknownClass.unknownMethod(){ |x| 1.type_assert(Object) }";
  ])

let suite = "suite">:::(success_suite @ failure_suite)

let run _  =
  if trace
  then (Cresolve.trace :=
	  fun cs -> print_string (printer cs); print_string "\n")
  else ();
  run_test_tt_main suite

let _ = Config.initialize None

let _ = run ()

