module Constr = struct

  type t  =
    | Less_than of Ikra_type.t * Ikra_type.t
    | Equal of Ikra_type.t * Ikra_type.t
    (* Has_method(tr,m,Tproc([t1;t2;...],t0,Some(Tproc([s1;s2;...],s0,None)))
       represents an existence of a method like:

       class tr
         def m(x1:t1, x2:t2, ...)
	   ... yield(y1:s1, y2:s2, ...):s0 ...
           v:t0
         end
       end 
    *)
    | Has_method of Ikra_type.t * string * Ikra_type.t
    (* class C def m(x1,..,xN) ... end end

       will create a constraint Defines_method(C, "m", tM) where
       tM = t1*..*tN -> t0 & tB, ti is type of xi, t0 is the return type,
       tB is the block type.
    *)
    | Defines_method of Ikra_type.t * string * Ikra_type.t

  let equal(t1,t2) = Equal(t1,t2)
  let less_than(t1,t2) = Less_than(t1,t2)
  let has_method(t1,m,t2) = Has_method(t1,m,t2)
  let defines_method(t1,m,t2) = Defines_method(t1,m,t2)

  let type_order = let open Ikra_type in function
    | Tint -> 0
    | Tfloat -> 1
    | Tbool -> 2
    | Tarray _ -> 3
    | Tobject _ -> 4
    | Tvar _ -> 5
    | Tclass _ -> 6
    | Ttuple _ -> 7
    | Tneighbor _ -> 8
    | Trange -> 9
    | Tproc _ -> 10

  let rec compare_types x y =
    let rec list_compare = function
      | ([],[])   -> 0
      | ([],_::_) -> -1
      | (_::_,[]) -> 1
      | (h1::t1,h2::t2) -> let c = compare_types h1 h2 in
                           if c=0 then list_compare (t1,t2)
                           else c in
    let open Ikra_type in match x,y with
    | x,y when x=y -> 0
    | x,y when type_order x < type_order y -> -1
    | x,y when type_order x > type_order y -> 1
    | Tarray(dim1,element1),Tarray(dim2,element2) when dim1=dim2 ->
      compare element1 element2
    | Tarray(dim1,element1),Tarray(dim2,element2) -> dim1 - dim2
    | Tvar(n1),Tvar(n2) -> n1 - n2
    | Tclass(c1),Tclass(c2) -> String.compare c1 c2
    | Ttuple(t1),Ttuple(t2) -> list_compare (t1,t2)
    | Tneighbor(m1,n1,t1),Tneighbor(m2,n2,t2) when m1=m2 && n1=n2 ->
      compare t1 t2
    | Tneighbor(m1,n1,t1),Tneighbor(m2,n2,t2) when m1=m2 -> n1 - n2
    | Tneighbor(m1,n1,t1),Tneighbor(m2,n2,t2) -> m1 - m2
    | Tproc(a1,r1,b1), Tproc(a2,r2,b2)
      when list_compare(a1,a2)=0 && r1=r2 -> (match (b1,b2) with
        | (Some bb1, Some bb2) -> compare_types bb1 bb2
        | (Some _,   None)     -> -1
        | (None,     Some _)   -> 1
        | (None,     None)     -> 0)
    | Tproc(a1,r1,_), Tproc(a2,r2,_)
      when list_compare(a1,a2)=0 -> compare_types r1 r2
    | Tproc(a1,r1,_   ), Tproc(a2,r2,_)
      -> list_compare(a1,a2)
    | _ -> invalid_arg ("dummy fallthrough: x="^ (string_of x) ^", y="^
                           (string_of y))


  let rec compare_type_list = function	(* dictionary order *)
    | [],[] -> 0
    | [],_  -> 1
    | _,[]  -> -1
    | t1::r1, t2::r2 when compare_types t1 t2 = 0
		     -> compare_type_list(r1,r2)
    | t1::r1, t2::r2 -> compare_types t1 t2

  let compare rel1 rel2 =
    let rank = function
      | Less_than _      -> 1
      | Equal _          -> 2
      | Has_method _     -> 3
      | Defines_method _ -> 4 in
    (* let maybe_subtypes = function *)
    (*   | None      -> [] *)
    (*   | Some(_,t) -> [t] in *)
    let subtypes = function
      | Less_than(t1,t2) | Equal(t1,t2) -> [t1;t2]
      | Has_method(tr,m,tp)             -> [tr;tp]
      | Defines_method(tr,m,tp)         -> [tr;tp] in

    let d = (rank rel1) - (rank rel2) in
    if d <> 0 then d
    else match (rel1,rel2) with
      | (Less_than _,_) 
      | (Equal _,_) -> 
	compare_type_list (subtypes rel1, subtypes rel2)
      | (Has_method(_,m1,_),Has_method(_,m2,_))
	  when String.compare m1 m2 = 0 -> 
	compare_type_list (subtypes rel1, subtypes rel2)
      | (Has_method(_,m1,_),Has_method(_,m2,_)) 
	-> String.compare m1 m2
      | (Defines_method(_,m1,_),Defines_method(_,m2,_))
          when String.compare m1 m2 = 0 ->
        compare_type_list (subtypes rel1, subtypes rel2)
      | (Defines_method(_,m1,_),Defines_method(_,m2,_)) 
	-> String.compare m1 m2
      | _ -> invalid_arg "d<>0 but rel1 or rel2 is not Has_method"

  let cmp r1 r2 = compare r1 r2 = 0

  let string_of =
    (* let maybe_string_of_block = function *)
    (*   | Some(ss,s0) -> Printf.sprintf ",&(%s->%s)" *)
    (*     (String.concat "," (List.map Ikra_type.string_of ss)) *)
    (*     (Ikra_type.string_of s0) *)
    (*   | None -> "" in *)
    function
    | Less_than(t1,t2) ->
      (Ikra_type.string_of t1) ^ "<:" ^ (Ikra_type.string_of t2)
    | Equal(t1,t2) ->
      (Ikra_type.string_of t1) ^ "=" ^ (Ikra_type.string_of t2)
    | Has_method(tr,m,tp) ->
      Printf.sprintf "mtype(%s,%s)= %s"
	(Ikra_type.string_of tr) m (Ikra_type.string_of tp)
    | Defines_method(tr,m,tp) ->
      Printf.sprintf "def(%s,%s)= %s"
	(Ikra_type.string_of tr) m (Ikra_type.string_of tp) 
end

exception Contradiction of Constr.t

module Constraints = struct

let check : Constr.t -> Constr.t =
  (* check c returns c as it is after checking whether c is
     not contradicting.  If contradicting, it raises a
     Contradiction exception *)
  let open Constr in
  let open Ikra_type in
  fun c -> match c with
    | Equal(Tobject,Tint)
    | Equal(Tint,Tobject)
    | Equal(Tobject,Tbool)
    | Equal(Tbool,Tobject)
      -> raise (Contradiction c)
    | _ -> c

let add_check : ('a -> Constr.t) -> 'a -> Constr.t =
  fun gen x -> check (gen x)


  module C = Set.Make(Constr)
  type elt = C.elt
  type t = C.t
  let empty = C.empty
  let is_empty = C.is_empty
  let mem = C.mem
  let add elt set = C.add (check elt) set
  let singleton = C.singleton
  let remove = C.remove
  let union = C.union
  let inter = C.inter
  let diff = C.diff
  let compare = C.compare
  let equal = C.equal
  let subset = C.subset
  let iter = C.iter
  let fold = C.fold
  let for_all = C.for_all
  let exists = C.exists
  let filter = C.filter
  let partition = C.partition
  let cardinal = C.cardinal
  let elements = C.elements
  let min_elt = C.min_elt
  let max_elt = C.max_elt
  let choose = C.choose
  let split = C.split

end 

let set_print = fun s -> "{" ^ 
  (String.concat "," (List.map Constr.string_of
                        (Constraints.elements s))) ^ 
  "}"

let remove_method_defs : Constraints.t -> Constraints.t =
  Constraints.filter (function
    | Constr.Defines_method _ -> false
    | _ -> true)

