open Ast

let maybe = Util.Option.maybe

let recognize_decls : tast list -> type_decl list =

  (* to convert given the type declaration in the source code to a
     list of type declarations

     recognize_decls (parse "(Integer, :a, :b, Float, :c)") 
     -> ["a", Tint; "b", Tint; "c", Tfloat]
  *)

  fun decls -> 
    let rec loop = function
      | (_,[]) -> []
      | (Some previous_type, (Literal(Symbol_literal(variable)),_)::rest) -> 
        (variable,previous_type)::(loop (Some previous_type, rest))
      | (_, texp::(Literal(Symbol_literal(variable)),_)::rest)
          when Ast.is_ikra_type texp ->
        let new_type = Ast.ikra_type_of_ast texp in
        (variable, new_type)::(loop (Some new_type, rest))

      | _ -> raise (Invalid_argument
                      ("ill formated type decl: " ^
                          (String.concat "," (List.map (fun (e) -> Ast.string_of e) decls))))
    in loop (None, decls)


let rec recognize_type_decls : tast -> tast =
  
  (* to construct, given the source AST in which type declarations are
     embedded as pseudo-method calls, the same AST except for those
     type declarations are represented special AST nodes.

     Example:

     recognize_type_decls
     (Block(None, [Call(None, "_typedecl",
     [Const("Integer"),
     Literal(Symbol("a")), Literal(Symbol("b"))],
     None);
     Local_assign("a",Local_var("b"))]))
     = Block(Ast.empty_pattern,
     [Type_decl(["a","Integer"; "b","Integer"], 
     Local_assign("a",Local_var("b")))])

     which corresponds to an expression: "_typedecl(Integer, :a, :b); a=b"

  *)

  let recognize_and_replace
      : tast -> parameter_pattern -> (tast * parameter_pattern) =
    fun body pattern ->
      match recognize_type_decls body with
        | Type_decl(decls,dbody),dt ->
          (let lookup var default =
             try List.assoc var decls with Not_found -> default in
           let rec replace_types = function
             | Var_pattern(v,t) -> Var_pattern(v, lookup v t)
             | Array_pattern(pats) ->
               Array_pattern(List.map replace_types pats) in
           ((Type_decl(decls,dbody), lookup "_RETURN" dt),
            replace_types pattern))
        | _ -> invalid_arg "no typedecl in body" in

  function

    (* All _typedecl calls must be captured in its surrounding
       expression.  Hence the following case should not occur.  *)
    | Call(_, "_typedecl", _, _),_ ->
      raise (Invalid_argument "pseudo _typedecl call appears at an \
                               unexpected position.")

    (* method call *)
    | Call(receiver, method_name, operands, block),t ->
      Call(recognize_type_decls receiver,
           method_name,
           List.map recognize_type_decls operands,
           maybe recognize_type_decls block),t
    (* literal *)
    | Literal(literal),t -> Literal(literal),t
    (* assignment to a constant *)
    | Constant_assign(name,ast),t ->
      Constant_assign(name,recognize_type_decls ast),t
    (* constant  *)
    | Constant(name),t -> Constant(name),t
    (* assignment to local variable *)
    | Local_assign(name,ast),t -> Local_assign(name,recognize_type_decls ast),t
    (* local variable reference *)
    | Local_var(name),t -> Local_var(name),t

    (* block structure with a type_decl pseudo call *)
    | Block((Call(receiver, "_typedecl", decls, None),t1)::body),t
      when Ast.is_self(receiver) ->
      (* Printf.printf "body=%s\n" *)
      (*    (String.concat ";" *)
      (*     (List.map (Ast.mlexp_of Ikra_type.string_of) body)); *)
        let body' = List.map recognize_type_decls body in
        (Type_decl(recognize_decls decls, body'),t)

    (* (normal) block structure *)
    | Block(body),t ->
      Block(List.map recognize_type_decls body),t

    | Proc(variables,body),t ->
      let body', variables' = recognize_and_replace body variables in
      Proc(variables', body'),t

    (* method definition (method_name, variables, body) *)
    | Method_def(method_name, variables, body),t ->

      let body', variables' = recognize_and_replace body variables in
      Method_def(method_name, variables', body'),t

    (* yield (ast) *)
    | Yield(ast),t -> Yield(recognize_type_decls ast),t
    (* array literal (ast list) *)
    | Array asts,t -> Array (List.map recognize_type_decls asts),t
    (* while loop (cond, body) *)
    | While(cond, body),t -> While(recognize_type_decls cond,
                                   recognize_type_decls body),t
    (* conditional branch (cond, then, else) *)
    | If(cond_exp, then_exp, else_exp),t ->
      If(recognize_type_decls cond_exp,
         recognize_type_decls then_exp,
         maybe recognize_type_decls else_exp),t
    (* for loop (var, range, body) *)
    | For(var, range, body),t ->
      For(var, recognize_type_decls range, recognize_type_decls body),t

    (* return(exp1) *)
    (* TODO: should it recognize the return type specified in the
       type_decl? *)
    | Return(exp1),t -> Return(maybe recognize_type_decls exp1),t
      
    (* type_decl (tdecl list, body) --- only after type inference *)
    | Type_decl _,_ ->
      raise (Invalid_argument "already recognized")

(* let recognize_type_decls exp = *)
(*   match recognize_type_decls exp with *)
(*     | Block(_, [Type_decl(_,    [Block _,_]),_]),_ as exp' -> exp' *)
(*     | Block(p, [Type_decl(decls,body),tt]),tb -> *)
(*       Block(p, [Type_decl(decls,[Block(Array_pattern[],body),tt]),tt]),tb *)
(*     | exp -> exp *)

let counter = ref 0             (* the counter for type variables *)
let new_type_var () =                   (* create a new type variable *)
  let t = Ikra_type.Tvar !counter in
  counter := (!counter + 1); t 

let rec add_type_vars_to_parameter_pattern
    : Ast.parameter_pattern -> Ast.parameter_pattern = function
      | Var_pattern(v,Ikra_type.Tobject) ->
        Var_pattern(v,new_type_var())
      | Array_pattern(sub_patterns) ->
        Array_pattern(List.map add_type_vars_to_parameter_pattern sub_patterns)
      | _ -> invalid_arg "a variable pattern already has a type variable"

let add_type_vars : Ast.tast -> Ast.tast =

  (* to attach a type variable to each sub-expression in a given
     expression

     add_type_vars (If(Var("x",_), Const(0,_), Var("x",_)),_)
     -> (If(Var("x",t0), Const(0,t1), Var("x",t2)),t3)
  *)


  (* add a type variable *)
  let rec adder (exp,_) = (traverse exp, new_type_var ())
  (* recursively call adder on subexpressions *)
  and traverse = function
    | Call(receiver,name,args,block) ->
      Call(adder receiver, name, List.map adder args,
           Util.Option.maybe adder block)
    | Local_assign(v,exp) -> Local_assign(v,adder exp)
    | Constant_assign(v,exp) -> Constant_assign(v,adder exp)
    | Block(bodies) ->
      Block(List.map adder bodies)
    | Proc(params, body) ->
      Proc(add_type_vars_to_parameter_pattern params, adder body)
    | Method_def(m,parames,body) ->
      Method_def(m,add_type_vars_to_parameter_pattern parames,adder body)
    | Yield(exp) -> Yield(adder exp)
    | Array(exps) -> Array(List.map adder exps)
    | While(cond,body) -> While(adder cond,adder body)
    | If(cond,then_exp, else_exp) ->
      If(adder cond, adder then_exp, Util.Option.maybe adder else_exp)
    | For(var,range,body) -> For(var,adder range,adder body)

    (* attach type variables to those in type_decl *)
    | Type_decl(tdecls, bodies) ->
      Type_decl(List.map (fun (var,_) -> (var, new_type_var())) tdecls,
                List.map adder bodies)
    | Return(exp1) -> Return(Util.Option.maybe adder exp1)

    | Local_var _ | Constant _ | Literal _ as leaf_exp -> leaf_exp in
  fun exp -> (counter := 0; adder exp)


type tenv = Ikra_type.t Env.env

let rec last_type = function
  | (_,t)::[] -> t
  | _::texps -> last_type texps
  | [] -> invalid_arg "no type in the empty list"

let type_of_literal =
  let open Ikra_type in
      function
        | Integer_literal _ -> Tint
        | Float_literal _ -> Tfloat
        | _ -> Tobject

let rec gather : tenv -> Ast.tast -> Type_constraints.Constraints.t =
  let (==) t1 t2 = Type_constraints.Constr.Equal(t1,t2) in
  let (<=) t1 t2 = Type_constraints.Constr.Less_than(t1,t2) in
  let (+) set const = Type_constraints.Constraints.add const set in
  let (@) = Type_constraints.Constraints.union in
  let has reciver m args return block
      = Type_constraints.Constr.Has_method
    (reciver,m, Ikra_type.Tproc(args,return,block)) in
  let defs reciver m args return block
      = Type_constraints.Constr.Defines_method
    (reciver,m,Ikra_type.Tproc(args,return,block)) in
  let types = List.map (fun (_,t) -> t) in

  let rec type_var_env_of_block_pattern : parameter_pattern -> 'a Env.env =
    let rec traverse env = function
      | Var_pattern(v,t) -> Env.extend env [v,t]
      | Array_pattern(pats) -> List.fold_left traverse env pats in
    traverse Env.empty in

  let rec block_parameter_types
      : parameter_pattern
      -> (Ikra_type.t list * Type_constraints.Constraints.t) =
        (* to costruct a type scheme from a given block parameter
           pattern, along with a set of constraints.

           block_parameter_types | (x:tx,y:ty),tz |
           = [Ttuple[tx;ty]; tz] with tnew = Ttuple[tx;ty]
        *)
    let open Type_constraints.Constraints in function
        | Ast.Var_pattern (_,block_param_type) -> ([block_param_type], empty)
        | Ast.Array_pattern pats ->
          List.fold_right (function
            | Ast.Var_pattern (_,t)
              -> fun (typs,raints) -> (t::typs,raints)
            | Ast.Array_pattern sub_pats
              -> let sub_typs, sub_raints =
                   List.split(List.map block_parameter_types sub_pats) in
                 fun (typs,raints) -> 
                   ((Ikra_type.Ttuple (List.flatten sub_typs))::typs,
                    (List.fold_left union raints sub_raints))
              (* TODO: to support nested parameters, we need to change
                 The Ast.Var_pattern to include type of the array
                 pattern itself *)
              (* invalid_arg "nested pattern is not supported yet" *))
             pats ([],empty) in

  let unify_block_tdecl_patterns
      : parameter_pattern -> type_decl list -> Type_constraints.Constraints.t =
    (* given a parameter pattern of a block and a list of type
       declarations in the body of the block, generate constraints
       that the types of the same variable are equal.  Note that the
       list of type declarations can have variables not specified as
       the block parameter and vice versa.

       Example: when a block is { |x:t0,y:t1| _tdecl(x:t2,z:t3); ... },
       it generates a constraint t0 = t2.
    *)
    fun block_pattern -> 
      let tenv = type_var_env_of_block_pattern block_pattern in
      List.fold_left (fun craints (var,tdecl_param_type) ->
        try craints + (Env.lookup tenv var == tdecl_param_type)
        with Env.Unbound _ -> craints
      )
        Type_constraints.Constraints.empty in

  let return_type_pesudo_var = "$ return type $" in
  let extend_return_type : tenv -> Ikra_type.t -> tenv =
    fun env typ -> Env.extend env [return_type_pesudo_var, typ] in
  let return_type_in_env : tenv -> Ikra_type.t =
    fun env -> Env.lookup env return_type_pesudo_var in

  fun env ((exp,typ) as texp)  -> match exp with

    (*  A method call with a block is valid, if the receiver type
        has a method m whose formal parameter types are supertypes of 
        the types of actual parameters, whose return type is the
        type of the call expression, and whose block type has 
        m input parameter types and one return type parameter.
        
        G, vi:ui |- fj : rj
        G |- ei : ti   e0 hasMethod(m, s1*..*sn * ((u1*..*um)->rj) -> ta)
        -----------------------------------------------------------------
        G |- e0.m(e1,..,en){ |v1,..,vm| f1;...fk } : ta



        By using a proc type, we can reformulate the above rule in
        this way:
        
                    G |- ei : ti     G |- blk : tb
               G |- e0 hasMethod(m,ti*..*tn -> ta & tb)
               ----------------------------------------
                     G |- e0.m(e1,..,en)blk : ta

        Note that the method parameter types are the types of the
        arguments, but not their subtypes.  This is because the
        subsumption between formal and actual parameters will be 
        checked at the constraint resolution stage where hasMethod
        and defMethod constraints match.


                           G |- None : None


                          G,vi:ti |- e : t0
                --------------------------------------
                G |- { |v1,..,vm| e } : t1*..*tn -> t0
        

    *)
    | Call(((_,tr) as receiver),mname,args, blk) ->

      let actual_param_types = types args in
      let block_type = Util.Option.maybe (fun (_,t)->t) blk in
      let return_type = typ in
      let block_consraints = match blk with
        |Some b -> (gather env b)
        |None   -> Type_constraints.Constraints.empty in

      (gather env receiver) @ (map_gather env args) @ block_consraints +
        (has tr mname actual_param_types return_type block_type)

    (* THIS IS AN OLD DEFINTION.  KEEP IT UNTIL WE FINISH THE NEW ONE. *)

    (* | Call(((_,tr) as receiver),mname,args, blk) -> *)
    (*   (\* add constraints between formal and actual parameters *\) *)
    (*   let (param_types,param_constraints) =  *)
    (*     List.fold_right (fun argt (pts,set) -> *)
    (*       let pt = new_type_var () in (pt::pts, set + (argt <= pt))) *)
    (*       (types args) ([],Type_constraints.Constraints.empty) in *)

    (*   (\* constraints from a block argument; generates no constraint if *)
    (*      the call has no block argument. *\) *)
    (*   let (block_constraints,block_type) = *)
    (*     match blk with *)
    (*       |Some (Block(block_param, *)
    (*                    [Type_decl(tdecl_pat,_),block_body_t as tdecl]),btyp) -> *)
    (*         let p_types, p_raints = block_parameter_types block_param in *)
    (*         (\* block constraints from block parameters *\) *)
    (*         (unify_block_tdecl_patterns block_param tdecl_pat)  *)
    (*         (\* from block body *\) *)
    (*         @ (gather env tdecl) *)
    (*         (\* and from block parameters (in case of tuples) *\) *)
    (*         @ p_raints, *)
    (*         (\* block type *\) *)
    (*         Some(Ikra_type.Tproc(p_types, block_body_t, None)) *)
    (*         (\* Some(p_types, block_body_t) *\) *)
    (*       |None -> (Type_constraints.Constraints.empty, None) *)
    (*       |_ -> invalid_arg "AST node other than a block appears as a \ *)
    (*                          block parameter to a method call." *)
    (*   in *)

    (*   (gather env receiver) @ (map_gather env args) @ param_constraints  *)
    (*   @ block_constraints + *)
    (*     (has tr mname param_types typ (block_type)) *)

    (* END OF THE OLD DEFINITION. *)

    (* NOTE: a symbol can have a type of objects or a type of
       procedure.  When it is used as a block parameter like
       a.reduce(&:+), the plus symbol represents a block { |x,y| x+y}.
       To achieve this kind of polymorphism, a symbol literal itself
       does not generate consraints.  At the third stage of constraint
       resolution, all unresolved type variables are given object
       type.*)
    | Literal(Ast.Symbol_literal l) -> Type_constraints.Constraints.empty

    | Literal(l) ->
      Type_constraints.Constraints.singleton (typ ==  type_of_literal l)
    | Constant_assign(var,(e0,t0))
    | Local_assign(var,(e0,t0)) ->
      let t1 = Env.lookup env var in
      (gather env (e0,t0)) + (t0 <= t1) + (typ == t1)
    | Constant var when not(Env.is_bound env var) ->
      Type_constraints.Constraints.singleton (typ == Ikra_type.Tobject)
    | Local_var var | Constant var ->
      Type_constraints.Constraints.singleton (typ == Env.lookup env var)

    (* NOTE: it is incorrect to use a Block node for both Algol
       begin/end block structures as well as an anonymous block
       parameter.  The type of a (former) block should be the type of
       the last statement of the block.  The type of a latter block
       should be a procedure type. For example,
       
       if true then print 1; 2 else 0 end
       
       is actually a 

       if true then BLOCK[print 1; 2] else 0 end

       Therefore, the block in the then-brach should be of type
       integer.  However,

       m() do print 1; 2 end, which is equal to
       m() { print 1; 2 }, and to
       m() BLOCK[ print 1; 2 ]

       means that the block is an argument to m, which should be of
       type "procedure from unit to integer".

       SOLUTION: we introduce a new node Proc for block arguments.
       This will cause modifications in many places, but we should
       do...

    *)

    | Block(bodies) ->
      (map_gather env bodies) + (last_type bodies <= typ)

    | Proc(proc_param, 
           (Type_decl(tdecl_pat,_), block_body_t as tdecl)) ->
      let p_types, p_raints = block_parameter_types proc_param in
      (* block constraints from block parameters *)
      (unify_block_tdecl_patterns proc_param tdecl_pat)
      (* from block body *)
      @ (gather env tdecl)
      (* and from block parameters (in case of tuples) *)
      @ p_raints
      (* block type *)
      + (typ == Ikra_type.Tproc(p_types, block_body_t, None))

    | Proc(Array_pattern[], ((_,body_t) as body)) ->
      (* from block body *)
      (gather env body)
      (* block type *)
      + (typ == Ikra_type.Tproc([], body_t, None))

    | Proc _ ->
      invalid_arg("unexpected body in Proc: " ^ (Ast.string_of texp))

    | Method_def(mname, parameters, ((_,t) as body)) ->
      (* let ret_typ = new_type_var() in *)
      let ret_typ = t in
      let param_types_alist : (string * Ikra_type.t) list =
        match parameters with
          | Ast.Var_pattern (v,t) -> [v,t]
          | Ast.Array_pattern ps ->
            List.map (function Var_pattern(v,t) -> (v,t)
              | _ -> invalid_arg "no support for nested pattern")
              ps in
      let param_types = List.map (fun (_,t) -> t) param_types_alist in
      (gather (Env.extend (extend_return_type env ret_typ) param_types_alist)
         body)
      + (defs (Env.lookup env "self") mname param_types ret_typ None)
      (* + (t <= ret_typ) *)

    (* | Yield exp -> *)
    (* | Array exps -> *)

    | Array(exps) ->
      (map_gather env exps) +
        (typ ==(Ikra_type.Ttuple (List.map (fun (_,t)->t) exps)))

    | While((_,t0) as cond, ((_,t1) as body)) ->
      (gather env cond) @ (gather env body)
      + (t0 <= Ikra_type.Tbool) + (typ == Ikra_type.Tbool)
    | If(((_,t1) as cond),((_,t2) as then_exp),(Some ((_,t3) as else_exp))) ->
      (gather env cond) @ (gather env then_exp) @ (gather env else_exp)
      + (t1 <= Ikra_type.Tbool) + (t2 <= typ) + (t3 <= typ)

    (* NOTE: an IF-THEN-without-ELSE expression is of type Object *)
    | If(((_,t1) as cond),then_exp,None) ->
      (gather env cond) @ (gather env then_exp)
      + (typ == Ikra_type.Tobject) + (t1 <= Ikra_type.Tbool)

    | For(var, ((range,t_range) as r), body) ->
      (gather env r)
      @ (gather (Env.extend env [var,Ikra_type.Tint]) body)
      + (t_range <= Ikra_type.Trange)
      + ((Env.lookup env var) == Ikra_type.Tint)
        (* NOTE: for x in r do e; end returns the value of r; henc the
           type of the for statement is that of the range expression *)
      + (typ == t_range)

    | Type_decl(tdecls,bodies) ->
      (* print_string ("\ngather "^ (Ast.string_of texp)); *)
      (map_gather (Env.extend env tdecls) bodies)
      + (last_type bodies == typ)

    | Return(Some(exp1,typ1)) ->
      (gather env (exp1,typ1)) + (typ1 <= (return_type_in_env env))

    | Return(None) ->
      Type_constraints.Constraints.singleton
        ((return_type_in_env env) == Ikra_type.Tobject)

      

    | Array _ | Yield _ ->
      invalid_arg ("unimplemented type of expression:" ^
                      (Ast.string_of texp))

and map_gather : tenv -> Ast.tast list -> Type_constraints.Constraints.t =
  fun env -> List.fold_left (fun c exp ->
    Type_constraints.Constraints.union (gather env exp) c) Type_constraints.Constraints.empty
and maybe_gather : tenv -> Ast.tast option -> Type_constraints.Constraints.t =
  fun env -> function
    | Some exp -> gather env exp
    | None -> Type_constraints.Constraints.empty

(* initial environment *)
let init_env : Ikra_type.t Env.env =
  let open Ikra_type in
      Env.extend Env.empty
        ( (List.map (fun cname -> (cname, Tclass cname)) 
             ["Array";"Math";"Time";])
          @ ["self", Tobject;
             Ast.implicit_self_name, Tobject;
             "Math::PI", Tfloat;
             "Math::E", Tfloat;
             "ARGV", Tobject;
             "true", Tbool;
             "false", Tbool;
             "$0", Tobject;
             "__FILE__", Tobject;
            ])

let generate_constraints : Ast.tast -> Type_constraints.Constraints.t =
  gather init_env 
