#!/usr/bin/env ruby

require("optparse")
require("test/unit")
require("fileutils")

# A test program for Ikra.
#
# Requirements:
#
# * Test programs will be taken from practical programs.
#
# * Test by comparison: one program will be executed by pure Ruby,
#   Ikra with C++ backend, and (if available) Ikra with CUDA backend.
#   The results from the execution by Ikra will be compared to the
#   ones by pure Ruby.
#
# * Scaling test: one program will be executed with larger sizes with
#   a larger backend.  In such a case, the results should be checked
#   without running on pure Ruby.
#
# * Test suites are managed.  For example, we want to know that all
#   the tests are succeeded.  (Hence it should show the number of
#   succeeded and failed tests at the end.)

# Implementation

# load .ikra.conf if available
['.',ENV['HOME']].find do |d|
  p=File.expand_path('.ikra.conf',d);
  File.exists?(p) && load(p)
end

IKRAC=File.expand_path("bin/ikrac", IKRA_HOME)
IKRASIM=File.expand_path("src/ikrasim", IKRA_HOME)

IKRA_GEN_DIR_NAME="ikra.gen"

# The next three lines interpret command-line arguments.  They should
# be done before 
$OPTIONS = ARGV.getopts('', 'cpp', 'cuda', 'rcuda')
$TESTS = ARGV.shift(ARGV.size)

class Backend
  attr_reader :name
  def initialize(name, executable, options)
    @name = name
    @executable = executable
    @options = options
  end
  def run(filename, parameters)
    `#{@executable} #{@options} #{filename} #{parameters}`
  end
end

class RemoteBackend < Backend
  def sys(command)
#    puts "PWD=#{Dir.pwd}; #{command}"
    `#{command}`
  end
  def remote_dir()
    /^.*?:(.*)/.match(IKRA_REMOTE_DIR) && $1
  end

  def setup(filename)
    @dirname=File.dirname(filename)
    @basename=File.basename(filename, ".rb")
  end

  def build(filename)
    if @last_build != filename  # to avoid building again when the
                                # same program is executed with a
                                # different parameter.
      @last_build = filename
      gen_dir = File.expand_path(IKRA_GEN_DIR_NAME, @dirname)
      if Dir.exist?(gen_dir)
        FileUtils.rm_r(gen_dir)
      end
      FileUtils.mkdir(gen_dir)
      sys("#{IKRAC} --gen --target-dir #{gen_dir} #{@options} #{filename}")
      sys("rsync -av #{@dirname} #{IKRA_REMOTE_DIR}")
      sys("ssh #{IKRA_REMOTE_HOST} '. .ikra.conf; cd #{remote_dir}/#{@dirname}/#{IKRA_GEN_DIR_NAME}; sh build.sh --no-exec'")
    end
  end

  def exec(filename, parameters)
    sys("ssh #{IKRA_REMOTE_HOST} '. .ikra.conf; cd #{remote_dir}/#{@dirname}; ${RUBY:-ruby} #{IKRA_GEN_DIR_NAME}/#{@basename}.ikra.rb #{parameters}'")
  end

  def run(filename, parameters)
    setup(filename)
    build(filename)
    exec(filename,parameters)
  end
end

PURE_RUBY_BACKEND = Backend.new("ruby", "ruby","")
IKRASIM_BACKEND = Backend.new("ikrasim", IKRASIM,"")
BACKENDS = {
  "cpp"   => Backend.new("C++", IKRAC,"--backend cpp"),
  "cuda"  => Backend.new("CUDA", IKRAC,"--backend cuda"),
  "rcuda" =>RemoteBackend.new("remote CUDA", IKRAC,"--backend cuda")
}

CUSTOM_ASSERT = "custom_assert.rb"

class Tests < Test::Unit::TestCase
  def targets()
    Dir.glob("*/*.rb").select{ |path|
      /(.*?)\/\1.rb/.match(path)
    }
  end
  def parameters(test_prog, scale)
    param_file = #File.expand_path(
      scale+".param"#, File.dirname(test_prog))
    out_file = #File.expand_path(
      scale+".out"#, File.dirname(test_prog))
    open(param_file).readlines.zip(File.exist?(out_file) ?
                                   open(out_file).readlines.map{ | line |
                                     eval(line)
                                   } :
                                   [] )
  end

  def backends()
    # if no backends are specified, select all of them
    ($OPTIONS.values.any? ?
     BACKENDS.each_pair.select{ |name, backend|
       $OPTIONS.empty? || $OPTIONS[name]
     } : BACKENDS ).
      map{ |_, backend| backend }
  end

  def run_with_backend(backend)
    test_dir = Dir.pwd
    targets().each do |test_prog|
      if $TESTS==[] || !$TESTS.include?(File.dirname(test_prog))
#       puts "skipping #{File.dirname(test_prog)}"
        next
      end
      print "testing  #{File.dirname(test_prog)}(#{backend.name})";
      STDOUT.flush
      tested="\tskipped\n"
      Dir.chdir(File.dirname(test_prog)) # 
      custom_assert_def = File.expand_path(CUSTOM_ASSERT,
                                           "." # File.dirname(test_prog)
                                           )
      if use_custom_assert = File.exists?(custom_assert_def)
        load(custom_assert_def)
      end
      scale="small"
      parameters(test_prog, scale).each do |param, out|
        print "."; STDOUT.flush; tested="\tdone\n"
        p = param.chomp
        expected = out || IKRASIM_BACKEND.run(File.basename(test_prog), p)
        actual = backend.run(File.basename(test_prog), p)
        msg = "Executions of #{test_prog} #{p} does not match."
        if use_custom_assert
          custom_assert(expected, actual, msg)
        else
          assert_equal(expected, actual, msg)
        end
      end
      Dir.chdir(test_dir)
      print tested; STDOUT.flush
    end
  end

  def test_ikra()
    backends().each do |backend|
      run_with_backend(backend)
    end
  end
end

